import os,sys,json,glob,shutil
import dpdata
import numpy as np
import unittest
import importlib

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'generator'
from .context import make_fp
from .context import detect_multiplicity
from .context import parse_cur_job
from .context import param_file
from .context import param_old_file
from .context import param_pwscf_file
from .context import param_pwscf_old_file
from .context import param_siesta_file
from .context import param_gaussian_file
from .context import param_cp2k_file
from .context import param_cp2k_file_v1
from .context import machine_file
from .context import param_diy_file
from .context import make_kspacing_kpoints
from .context import my_file_cmp
from .context import setUpModule
from .comp_sys import test_atom_names
from .comp_sys import test_atom_types
from .comp_sys import test_coord
from .comp_sys import test_cell
from pymatgen.io.vasp import Kpoints,Incar
from .context import param_pwmat_file
import scipy.constants as pc

vasp_incar_ref = "PREC=A\n\
ENCUT=600\n\
ISYM=0\n\
ALGO=fast\n\
EDIFF=1e-05\n\
LREAL=A\n\
NPAR=1\n\
KPAR=1\n\
NELMIN=4\n\
ISIF=2\n\
ISMEAR=1\n\
SIGMA=0.25\n\
IBRION=-1\n\
NSW=0\n\
LWAVE=F\n\
LCHARG=F\n\
PSTRESS=0\n\
KSPACING=0.16\n\
KGAMMA=F\n";

vasp_incar_ele_temp_ref = "PREC=A\n\
ENCUT=600\n\
ISYM=0\n\
ALGO=fast\n\
EDIFF=1e-05\n\
LREAL=A\n\
NPAR=1\n\
KPAR=1\n\
NELMIN=4\n\
ISIF=2\n\
ISMEAR=-1\n\
SIGMA=%.10f\n\
IBRION=-1\n\
NSW=0\n\
LWAVE=F\n\
LCHARG=F\n\
PSTRESS=0\n\
KSPACING=0.16\n\
KGAMMA=F\n";

pwscf_input_ref="&control\n\
calculation='scf',\n\
restart_mode='from_scratch',\n\
outdir='./OUT',\n\
tprnfor=.TRUE.,\n\
tstress=.TRUE.,\n\
disk_io='none',\n\
pseudo_dir='./',\n\
/\n\
&system\n\
vdw_corr='TS',\n\
ecutwfc=110,\n\
ts_vdw_econv_thr=1e-08,\n\
nosym=.TRUE.,\n\
ibrav=0,\n\
nat=6,\n\
ntyp=3,\n\
/\n\
&electrons\n\
conv_thr=1e-08,\n\
/\n"

siesta_input_ref="\
SystemName        system\n\
SystemLabel       system\n\
NumberOfAtoms     6\n\
NumberOfSpecies   3\n\
\n\
WriteForces       T\n\
WriteCoorStep     T\n\
WriteCoorXmol     T\n\
WriteMDXmol       T\n\
WriteMDHistory    T\n\
\n\
MeshCutoff            300 Ry\n\
DM.Tolerance          1.000000e-04\n\
DM.NumberPulay         5\n\
DM.UseSaveDM          true\n\
XC.functional          GGA\n\
XC.authors             PBE\n\
MD.UseSaveXV           T\n\
\n\
DM.UseSaveDM           F\n\
WriteDM                F\n\
WriteDM.NetCDF         F\n\
WriteDMHS.NetCDF       F\n\
\n\
%block Chemical_Species_label\n\
1	6	C\n\
2	1	H\n\
3	7	N\n"


gaussian_input_ref="""%nproc=14
#force b3lyp/6-31g*

DPGEN
"""

cp2k_input_ref="\
&GLOBAL\n\
PROJECT DPGEN\n\
&END GLOBAL\n\
&FORCE_EVAL\n\
METHOD QS\n\
STRESS_TENSOR ANALYTICAL\n\
&DFT\n\
BASIS_SET_FILE_NAME ./cp2k_basis_pp_file/BASIS_MOLOPT\n\
POTENTIAL_FILE_NAME ./cp2k_basis_pp_file/GTH_POTENTIALS\n\
CHARGE 0\n\
UKS F\n\
MULTIPLICITY 1\n\
&MGRID\n\
CUTOFF 400\n\
REL_CUTOFF 50\n\
NGRIDS 4\n\
&END MGRID\n\
&QS\n\
EPS_DEFAULT 1.0E-12\n\
&END QS\n\
&SCF\n\
SCF_GUESS ATOMIC\n\
EPS_SCF 1.0E-6\n\
MAX_SCF 50\n\
&OT\n\
MINIMIZER DIIS\n\
PRECONDITIONER FULL_SINGLE_INVERSE\n\
&END OT\n\
&END SCF\n\
&XC\n\
&XC_FUNCTIONAL PBE\n\
&END XC_FUNCTIONAL\n\
&END XC\n\
&END DFT\n\
&SUBSYS\n\
&CELL\n\
&END CELL\n\
&COORD\n\
@include coord.xyz\n\
&END COORD\n\
&KIND H\n\
BASIS_SET DZVP-MOLOPT-GTH\n\
POTENTIAL GTH-PBE-q1\n\
&END KIND\n\
&KIND C\n\
BASIS_SET DZVP-MOLOPT-GTH\n\
POTENTIAL GTH-PBE-q4\n\
&END KIND\n\
&KIND N\n\
BASIS_SET DZVP-MOLOPT-GTH\n\
POTENTIAL GTH-PBE-q5\n\
&END KIND\n\
&END SUBSYS\n\
&PRINT\n\
&FORCES ON\n\
&END FORCES\n\
&END PRINT\n\
&END FORCE_EVAL\n";

pwmat_input_ref = "4 1\n\
in.atom=atom.config\n\
ecut=50\n\
e_error=0.0001\n\
rho_error=0.0001\n\
scf_iter0_1=6 4 3 0.0000 0.025 2\n\
scf_iter0_2=94 4 3 1.0000 0.025 2\n\
mp_n123=147 57 39 0 0 0 2\n\
out.wg=F\n\
out.rho=F\n\
out.mlmd=T\n\
job=scf\n\
IN.PSP1 = C.SG15.PBE.UPF\n\
IN.PSP2 = H.SG15.PBE.UPF\n\
IN.PSP3 = N.SG15.PBE.UPF\n";




def _box2lmpbox(orig, box) :
    lohi = np.zeros([3,2])
    for dd in range(3) :
        lohi[dd][0] = orig[dd]
    tilt = np.zeros(3)
    tilt[0] = box[1][0]
    tilt[1] = box[2][0]
    tilt[2] = box[2][1]
    lens = np.zeros(3)
    lens[0] = box[0][0]
    lens[1] = box[1][1]
    lens[2] = box[2][2]
    for dd in range(3) :
        lohi[dd][1] = lohi[dd][0] + lens[dd]
    return lohi, tilt

def _box2dumpbox(orig, box) :
    lohi, tilt = _box2lmpbox(orig, box)
    xy = tilt[0]
    xz = tilt[1]
    yz = tilt[2]
    bounds = np.zeros([3,2])
    bounds[0][0] = lohi[0][0] + min(0.0,xy,xz,xy+xz)
    bounds[0][1] = lohi[0][1] + max(0.0,xy,xz,xy+xz)
    bounds[1][0] = lohi[1][0] + min(0.0,yz)
    bounds[1][1] = lohi[1][1] + max(0.0,yz)
    bounds[2][0] = lohi[2][0]
    bounds[2][1] = lohi[2][1]
    return bounds, tilt


def _write_lammps_dump(sys, dump_file, f_idx = 0) :
    cell = sys['cells'][f_idx].reshape([3,3])
    coord = sys['coords'][f_idx].reshape([-1,3])
    bd, tilt = _box2dumpbox(np.zeros(3), cell)
    atype = sys['atom_types']
    natoms = len(sys['atom_types'])
    with open(dump_file, 'w') as fp:
        fp.write('ITEM: TIMESTEP\n')
        fp.write('0\n')
        fp.write('ITEM: NUMBER OF ATOMS\n')
        fp.write(str(natoms) + '\n')
        fp.write('ITEM: BOX BOUNDS xy xz yz pp pp pp\n')
        for ii in range(3):
            fp.write('%f %f %f\n' % (bd[ii][0], bd[ii][1], tilt[ii]))
        fp.write('ITEM: ATOMS id type x y z\n')
        for ii in range(natoms) :
            fp.write('%d %d %f %f %f\n' % (ii+1, atype[ii]+1, coord[ii][0], coord[ii][1], coord[ii][2]))


def _make_fake_md(idx, md_descript, atom_types, type_map, ele_temp = None) :
    """
    md_descript: list of dimension
                 [n_sys][n_MD][n_frame]
    ele_temp: list of dimension
                 [n_sys][n_MD]
    """
    natoms = len(atom_types)
    ntypes = len(type_map)
    atom_types = np.array(atom_types, dtype = int)
    atom_numbs = [np.sum(atom_types == ii) for ii in range(ntypes)]
    sys = dpdata.System()
    sys.data['atom_names'] = type_map
    sys.data['atom_numbs'] = atom_numbs
    sys.data['atom_types'] = atom_types
    for sidx,ss in enumerate(md_descript) :
        for midx,mm in enumerate(ss) :
            nframes = len(mm)
            cells  = np.random.random([nframes,3,3])
            coords = np.random.random([nframes,natoms,3])
            sys.data['coords'] = coords
            sys.data['cells'] = cells
            task_dir = os.path.join('iter.%06d' % idx,
                                    '01.model_devi',
                                    'task.%03d.%06d' % (sidx, midx))
            os.makedirs(os.path.join(task_dir, 'traj'), exist_ok = True)
            for ii in range(nframes) :
                _write_lammps_dump(sys,
                                   os.path.join(task_dir,
                                                'traj',
                                                '%d.lammpstrj' % ii))
            md_out = np.zeros([nframes, 7])
            md_out[:,0] = np.arange(nframes)
            md_out[:,4] = mm
            np.savetxt(os.path.join(task_dir, 'model_devi.out'), md_out)
            if ele_temp is not None:
                with open(os.path.join(task_dir, 'job.json'), 'w') as fp:
                    json.dump({"ele_temp": ele_temp[sidx][midx]}, fp)


def _check_poscars(testCase, idx, fp_task_max, type_map) :
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    candi_files = glob.glob(os.path.join(fp_path, 'candidate.shuffled.*.out'))
    candi_files.sort()
    sys_idx = [str(os.path.basename(ii).split('.')[2]) for ii in candi_files]
    for sidx,ii in zip(sys_idx, candi_files) :
        md_task = []
        f_idx = []
        with open(ii) as fp:
            for ii in fp :
                md_task.append(ii.split()[0])
                f_idx.append(ii.split()[1])
        md_task = md_task[:fp_task_max]
        f_idx = f_idx[:fp_task_max]
        cc = 0
        for tt,ff in zip(md_task, f_idx) :
            traj_file = os.path.join(tt, 'traj', '%d.lammpstrj' % int(ff))
            poscar_file = os.path.join(fp_path,
                                       'task.%03d.%06d' % (int(sidx), cc),
                                       'POSCAR')
            cc += 1
            sys0 = dpdata.System(traj_file, fmt = 'lammps/dump', type_map = type_map)
            sys1 = dpdata.System(poscar_file, fmt = 'vasp/poscar')
            test_atom_names(testCase, sys0, sys1)

def _check_kpoints_exists(testCase, idx) :
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    for ii in tasks :
        testCase.assertTrue(os.path.isfile(os.path.join(ii, 'KPOINTS')))

def _check_kpoints(testCase, idx) :
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    for ii in tasks :
        kpoints=Kpoints.from_file(os.path.join(os.path.join(ii, 'KPOINTS')))
        incar=Incar.from_file(os.path.join(os.path.join(ii, 'INCAR')))
        kspacing = incar['KSPACING']
        gamma = incar['KGAMMA']
        if isinstance(gamma,bool):
           pass
        else:
           if gamma[0].upper()=="T":
              gamma=True
           else:
              gamma=False
        ret=make_kspacing_kpoints(os.path.join(os.path.join(ii, 'POSCAR')), kspacing, gamma)
        kpoints_ref=Kpoints.from_string(ret)
        testCase.assertEqual(repr(kpoints), repr(kpoints_ref))


def _check_incar_exists(testCase, idx) :
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    # testCase.assertTrue(os.path.isfile(os.path.join(fp_path, 'INCAR')))
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    for ii in tasks :
        my_file_cmp(testCase,
                    os.path.join(fp_path, 'INCAR'),
                    os.path.join(ii, 'INCAR'))


def _check_potcar(testCase, idx, fp_pp_path, fp_pp_files) :
    nfile = len(fp_pp_files)
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    for ii in range(nfile):
        testCase.assertTrue(os.path.isfile(os.path.join(fp_pp_path, fp_pp_files[ii])))
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    for ii in tasks :
        for jj in range(nfile):
            my_file_cmp(testCase,
                        os.path.join(fp_pp_path, fp_pp_files[jj]),
                        os.path.join(ii, fp_pp_files[jj]))


def _check_sel(testCase, idx, fp_task_max, flo, fhi):
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    candi_files = glob.glob(os.path.join(fp_path, 'candidate.shuffled.*.out'))
    candi_files.sort()
    sys_idx = [str(os.path.basename(ii).split('.')[2]) for ii in candi_files]
    for sidx,ii in zip(sys_idx, candi_files) :
        md_task = []
        f_idx = []
        with open(ii) as fp:
            for ii in fp :
                md_task.append(ii.split()[0])
                f_idx.append(ii.split()[1])
        md_task = md_task[:fp_task_max]
        f_idx = f_idx[:fp_task_max]
        for tt,ff in zip(md_task, f_idx):
            md_value = np.loadtxt(os.path.join(tt, 'model_devi.out'))
            fvalue = md_value[int(ff)][4]
            testCase.assertTrue(fvalue >= flo)
            testCase.assertTrue(fvalue <  fhi)


def _check_incar(testCase, idx):
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    cwd = os.getcwd()
    for ii in tasks :
        os.chdir(ii)
        with open('INCAR') as fp:
            incar = fp.read()
            testCase.assertEqual(incar.strip(), vasp_incar_ref.strip())
        os.chdir(cwd)

def _check_incar_ele_temp(testCase, idx, ele_temp):
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    cwd = os.getcwd()
    for ii in tasks :
        os.chdir(ii)
        bname = os.path.basename(ii)
        sidx = int(bname.split('.')[1])
        tidx = int(bname.split('.')[2])
        with open('INCAR') as fp:
            incar = fp.read()
            incar0 = Incar.from_string(incar)
            # make_fake_md: the frames in a system shares the same ele_temp
            incar1 = Incar.from_string(vasp_incar_ele_temp_ref%(ele_temp[sidx][0] * pc.Boltzmann / pc.electron_volt))
            for ii in incar0.keys():
                # skip checking nbands...
                if ii == 'NBANDS':
                    continue
                testCase.assertAlmostEqual(incar0[ii], incar1[ii], msg = 'key %s differ' % (ii), places = 5)
        os.chdir(cwd)

def _check_pwscf_input_head(testCase, idx) :
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    for ii in tasks :
        ifile = os.path.join(ii, 'input')
        testCase.assertTrue(os.path.isfile(ifile))
        with open(ifile) as fp:
            lines = fp.read().split('\n')
        for idx, jj in enumerate(lines) :
            if 'ATOMIC_SPECIES' in jj :
                break
        lines = lines[:idx]
        testCase.assertEqual(('\n'.join(lines)).strip(), pwscf_input_ref.strip())

def _check_siesta_input_head(testCase, idx) :
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    for ii in tasks :
        ifile = os.path.join(ii, 'input')
        testCase.assertTrue(os.path.isfile(ifile))
        with open(ifile) as fp:
            lines = fp.read().split('\n')
        for idx, jj in enumerate(lines) :
            if '%endblock Chemical_Species_label' in jj :
                break
        lines = lines[:idx]
        testCase.assertEqual(('\n'.join(lines)).strip(), siesta_input_ref.strip())


def _check_gaussian_input_head(testCase, idx) :
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    for ii in tasks :
        ifile = os.path.join(ii, 'input')
        testCase.assertTrue(os.path.isfile(ifile))
        with open(ifile) as fp:
            lines = fp.read().split('\n')
        for idx, jj in enumerate(lines) :
            if '0 1' in jj :
                break
        lines = lines[:idx]
        testCase.assertEqual(('\n'.join(lines)).strip(), gaussian_input_ref.strip())


def _check_cp2k_input_head(testCase, idx) :
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    for ii in tasks :
        ifile = os.path.join(ii, 'input.inp')
        testCase.assertTrue(os.path.isfile(ifile))
        with open(ifile) as fp:
            lines = fp.read().split('\n')
        for idx, jj in enumerate(lines) :
            if '&CELL' in jj :
                cell_start_idx = idx
            if '&END CELL' in jj :
                cell_end_idx = idx
        lines_check = lines[:cell_start_idx+1] + lines[cell_end_idx:]
        testCase.assertEqual(('\n'.join(lines_check)).strip(), cp2k_input_ref.strip())



def _check_pwmat_input(testCase, idx):
    fp_path = os.path.join('iter.%06d' % idx, '02.fp')
    tasks = glob.glob(os.path.join(fp_path, 'task.*'))
    cwd = os.getcwd()
    for ii in tasks :
        os.chdir(ii)
        os.system("sed -i '8c mp_n123=147 57 39 0 0 0 2' etot.input")
        with open('etot.input') as fp:
            lines = fp.read()
            testCase.assertEqual(lines.strip(), pwmat_input_ref.strip())
        os.chdir(cwd)

class TestMakeFPPwscf(unittest.TestCase):
    def test_make_fp_pwscf(self):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_pwscf_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        md_descript = []
        nsys = 2
        nmd = 3
        n_frame = 10
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
        atom_types = [0, 1, 2, 2, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        _check_pwscf_input_head(self, 0)
        _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')

    def test_make_fp_pwscf_old(self):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_pwscf_old_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        md_descript = []
        nsys = 2
        nmd = 3
        n_frame = 10
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
        atom_types = [0, 1, 2, 2, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        _check_pwscf_input_head(self, 0)
        _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')

class TestMakeFPSIESTA(unittest.TestCase):
    def test_make_fp_siesta(self):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_siesta_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        md_descript = []
        nsys = 2
        nmd = 3
        n_frame = 10
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
        atom_types = [0, 1, 2, 2, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        _check_siesta_input_head(self, 0)
        _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')

class TestMakeFPVasp(unittest.TestCase):
    def test_make_fp_vasp(self):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        md_descript = []
        nsys = 2
        nmd = 3
        n_frame = 10
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
        atom_types = [0, 1, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        # _check_incar_exists(self, 0)
        _check_incar(self, 0)
        _check_kpoints_exists(self, 0)
        _check_kpoints(self,0)
        # checked elsewhere
        # _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')

    def test_make_fp_vasp_old(self):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_old_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        md_descript = []
        nsys = 2
        nmd = 3
        n_frame = 10
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
        atom_types = [0, 1, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        # _check_incar_exists(self, 0)
        _check_incar(self, 0)
        _check_kpoints_exists(self, 0)
        _check_kpoints(self,0)
        # checked elsewhere
        # _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')

    def test_make_fp_vasp_less_sel(self):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        md_descript = []
        nsys = 1
        nmd = 1
        n_frame = 8
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
        atom_types = [0, 1, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        # _check_incar_exists(self, 0)
        _check_incar(self, 0)
        _check_kpoints_exists(self, 0)
        _check_kpoints(self,0)
        # checked elsewhere
        # _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')


    def test_make_fp_vasp_from_incar(self):
        ## Verify if user chooses to diy VASP INCAR totally.
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_diy_file, 'r') as fp :
            jdata = json.load (fp)
        fp.close()
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        fp.close()
        md_descript = []
        nsys = 2
        nmd = 3
        n_frame = 10
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
        atom_types = [0, 1, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        # _check_incar_exists(self, 0)
        _check_incar(self, 0)
        _check_kpoints_exists(self, 0)
        _check_kpoints(self,0)
        # checked elsewhere
        # _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')

    def test_make_fp_vasp_ele_temp(self):
        ## Verify if user chooses to diy VASP INCAR totally.
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_diy_file, 'r') as fp :
            jdata = json.load (fp)
        fp.close()
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        fp.close()
        md_descript = []
        ele_temp = []
        nsys = 2
        nmd = 3
        n_frame = 10
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
            ele_temp.append([np.random.random() * 100000] * nmd)
        atom_types = [0, 1, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map, ele_temp = ele_temp)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        _check_incar_ele_temp(self, 0, ele_temp)
        _check_kpoints_exists(self, 0)
        _check_kpoints(self,0)
        # checked elsewhere
        # _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')


class TestMakeFPGaussian(unittest.TestCase):
    def make_fp_gaussian(self, multiplicity="auto"):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_gaussian_file, 'r') as fp :
            jdata = json.load (fp)
        jdata['user_fp_params']['multiplicity'] = multiplicity
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        md_descript = []
        nsys = 2
        nmd = 3
        n_frame = 10
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
        atom_types = [0, 1, 2, 2, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        _check_gaussian_input_head(self, 0)
        _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')

    @unittest.skipIf(importlib.util.find_spec("openbabel") is None, "requires openbabel")
    def test_make_fp_gaussian(self):
        self.make_fp_gaussian()

    def test_make_fp_gaussian_multiplicity_one(self):
        self.make_fp_gaussian(multiplicity=1)

    def test_detect_multiplicity(self):
        # oxygen O2 3
        self._check_multiplicity(['O', 'O'], 3)
        # methane CH4 1
        self._check_multiplicity(['C', 'H', 'H', 'H', 'H'], 1)
        # CH3 2
        self._check_multiplicity(['C', 'H', 'H', 'H'], 2)
        # CH2 1
        self._check_multiplicity(['C', 'H', 'H'], 1)
        # CH 2
        self._check_multiplicity(['C', 'H'], 2)

    def _check_multiplicity(self, symbols, multiplicity):
        self.assertEqual(detect_multiplicity(np.array(symbols)), multiplicity)

class TestMakeFPCP2K(unittest.TestCase):
    def test_make_fp_cp2k(self):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_cp2k_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        md_descript = []
        nsys = 2
        nmd = 3
        n_frame = 10
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
        atom_types = [0, 1, 2, 2, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        _check_cp2k_input_head(self, 0)
        _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')


class TestMakeFPPWmat(unittest.TestCase):
    def test_make_fp_pwmat(self):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_pwmat_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        md_descript = []
        nsys = 2
        nmd = 3
        n_frame = 10
        for ii in range(nsys) :
            tmp = []
            for jj in range(nmd) :
                tmp.append(np.arange(0, 0.29, 0.29/10))
            md_descript.append(tmp)
        atom_types = [0, 1, 2, 2, 0, 1]
        type_map = jdata['type_map']
        _make_fake_md(0, md_descript, atom_types, type_map)
        make_fp(0, jdata, {})
        _check_sel(self, 0, jdata['fp_task_max'], jdata['model_devi_f_trust_lo'], jdata['model_devi_f_trust_hi'])
        _check_poscars(self, 0, jdata['fp_task_max'], jdata['type_map'])
        _check_pwmat_input(self, 0)
        _check_potcar(self, 0, jdata['fp_pp_path'], jdata['fp_pp_files'])
        shutil.rmtree('iter.000000')

if __name__ == '__main__':
    unittest.main()





#!/usr/bin/env python3 

import os
import re
import sys
import argparse
import glob
import json
import random
import logging
import warnings
import shutil
import time
import dpdata
import numpy as np
from dpgen import dlog
import subprocess as sp
import dpgen.data.tools.hcp as hcp
import dpgen.data.tools.fcc as fcc
import dpgen.data.tools.bcc as bcc
import dpgen.data.tools.diamond as diamond
import dpgen.data.tools.sc as sc
from dpgen.generator.lib.vasp import incar_upper
from pymatgen import Structure
from pymatgen.io.vasp import Incar
from dpgen.remote.decide_machine import  decide_fp_machine
from dpgen import ROOT_PATH
from dpgen.dispatcher.Dispatcher import Dispatcher, make_dispatcher



def create_path (path,back=False) :
    if  path[-1] != "/":
        path += '/'
    if os.path.isdir(path) : 
        if back:
           dirname = os.path.dirname(path)        
           counter = 0
           while True :
               bk_dirname = dirname + ".bk%03d" % counter
               if not os.path.isdir(bk_dirname) : 
                   shutil.move (dirname, bk_dirname) 
                   break
               counter += 1
           os.makedirs (path)
           return path
        else:
           return path

    os.makedirs (path)
    return path

def replace (file_name, pattern, subst) :
    file_handel = open (file_name, 'r')
    file_string = file_handel.read ()
    file_handel.close ()
    file_string = ( re.sub (pattern, subst, file_string) )
    file_handel = open (file_name, 'w')
    file_handel.write (file_string)
    file_handel.close ()

"""
0, make unit cell
1, copy
2, place element
3, relax
4, perturb
"""
global_dirname_02 = '00.place_ele'
global_dirname_03 = '01.scale_pert'
global_dirname_04 = '02.md'

def out_dir_name(jdata) :
    cell_type = jdata['cell_type']
    elements = jdata['elements']
    super_cell = jdata['super_cell']    
    from_poscar = False
    if 'from_poscar' in jdata :
        from_poscar = jdata['from_poscar']
        from_poscar_path = jdata['from_poscar_path']

    if from_poscar:
        poscar_name = os.path.basename(from_poscar_path)
        cell_str = "%02d" % (super_cell[0])
        for ii in range(1,len(super_cell)) :
            cell_str = cell_str + ("x%02d" % super_cell[ii])
        return poscar_name + '.' + cell_str
    else :
        ele_str = ""
        for ii in elements:
            ele_str = ele_str + ii.lower()
        cell_str = "%02d" % (super_cell[0])
        for ii in range(1,len(super_cell)) :
            cell_str = cell_str + ("x%02d" % super_cell[ii])
        return ele_str + '.' + cell_type + '.' + cell_str

def class_cell_type(jdata) :
    ct = jdata['cell_type']
    if ct == "hcp" :
        cell_type = hcp
    elif ct == "fcc" :
        cell_type = fcc
    elif ct == "diamond" :
        cell_type = diamond
    elif ct == "sc" :
        cell_type = sc
    elif ct == "bcc" :
        cell_type = bcc
    else :
        raise RuntimeError("unknown cell type %s" % ct)
    return cell_type

def poscar_ele(poscar_in, poscar_out, eles, natoms) :
    ele_line = ""
    natom_line = ""
    for ii in eles :
        ele_line += str(ii) + " "
    for ii in natoms :
        natom_line += str(ii) + " "
    with open(poscar_in, 'r') as fin :
        lines = list(fin)
        lines[5] = ele_line + "\n"
        lines[6] = natom_line + "\n"
    with open(poscar_out, 'w') as fout :
        fout.write("".join(lines))

def poscar_natoms(lines) :
    numb_atoms = 0
    for ii in lines[6].split() :
        numb_atoms += int(ii)
    return numb_atoms

def poscar_shuffle(poscar_in, poscar_out) :
    with open(poscar_in, 'r') as fin :
        lines = list(fin)
    numb_atoms = poscar_natoms(lines)
    idx = np.arange(8, 8+numb_atoms)
    np.random.shuffle(idx)
    out_lines = lines[0:8]
    for ii in range(numb_atoms) :
        out_lines.append(lines[idx[ii]])
    with open(poscar_out, 'w') as fout:
        fout.write("".join(out_lines))

def poscar_scale_direct (str_in, scale) :
    lines = str_in.copy()
    numb_atoms = poscar_natoms(lines)
    pscale = float(lines[1])
    pscale = pscale * scale
    lines[1] = str(pscale) + "\n"
    return lines

def poscar_scale_cartesian (str_in, scale) :
    lines = str_in.copy()
    numb_atoms = poscar_natoms(lines)
    # scale box
    for ii in range(2,5) :
        boxl = lines[ii].split()
        boxv = [float(ii) for ii in boxl]
        boxv = np.array(boxv) * scale
        lines[ii] = "%.16e %.16e %.16e\n" % (boxv[0], boxv[1], boxv[2])
    # scale coord
    for ii in range(8, 8+numb_atoms) :
        cl = lines[ii].split()
        cv = [float(ii) for ii in cl]
        cv = np.array(cv) * scale
        lines[ii] = "%.16e %.16e %.16e\n" % (cv[0], cv[1], cv[2])
    return lines    

def poscar_scale (poscar_in, poscar_out, scale) :
    with open(poscar_in, 'r') as fin :
        lines = list(fin)
    if 'D' == lines[7][0] or 'd' == lines[7][0]: 
        lines = poscar_scale_direct(lines, scale)
    elif 'C' == lines[7][0] or 'c' == lines[7][0] :
        lines = poscar_scale_cartesian(lines, scale)
    else :
        raise RuntimeError("Unknow poscar style at line 7: %s" % lines[7])
    with open(poscar_out, 'w') as fout:
        fout.write("".join(lines))

def make_unit_cell (jdata) :
    latt = jdata['latt']
    out_dir = jdata['out_dir']
    path_uc = os.path.join(out_dir, global_dirname_02)
    cell_type = class_cell_type(jdata)

    cwd = os.getcwd()    
    # for ii in scale :
    # path_work = create_path(os.path.join(path_uc, '%.3f' % ii))
    path_work = create_path(path_uc)    
    os.chdir(path_work)
    with open('POSCAR.unit', 'w') as fp:
        fp.write (cell_type.poscar_unit(latt))
    os.chdir(cwd)        

def make_super_cell (jdata) :
    out_dir = jdata['out_dir']
    super_cell = jdata['super_cell']
    path_uc = os.path.join(out_dir, global_dirname_02)
    path_sc = os.path.join(out_dir, global_dirname_02)
    assert(os.path.isdir(path_uc)), "path %s should exists" % path_uc
    assert(os.path.isdir(path_sc)), "path %s should exists" % path_sc

    # for ii in scale :
    from_path = path_uc
    from_file = os.path.join(from_path, 'POSCAR.unit')
    to_path = path_sc
    to_file = os.path.join(to_path, 'POSCAR')

    #minor bug for element symbol behind the coordinates
    from_struct=Structure.from_file(from_file)
    from_struct.make_supercell(super_cell)
    from_struct.to('poscar',to_file)

def make_super_cell_poscar(jdata) :
    out_dir = jdata['out_dir']
    super_cell = jdata['super_cell']
    path_sc = os.path.join(out_dir, global_dirname_02)    
    create_path(path_sc)
    from_poscar_path = jdata['from_poscar_path']
    assert(os.path.isfile(from_poscar_path)), "file %s should exists" % from_poscar_path
    
    from_file = os.path.join(path_sc, 'POSCAR.copied')
    shutil.copy2(from_poscar_path, from_file)
    to_path = path_sc
    to_file = os.path.join(to_path, 'POSCAR')
  
    #minor bug for element symbol behind the coordinates
    from_struct=Structure.from_file(from_file)
    from_struct.make_supercell(super_cell)
    from_struct.to('poscar',to_file)  

    # make system dir (copy)
    lines = open(to_file, 'r').read().split('\n')
    natoms_str = lines[6]
    natoms_list = [int(ii) for ii in natoms_str.split()]
    dlog.info(natoms_list)
    comb_name = "sys-"
    for idx,ii in enumerate(natoms_list) :
        comb_name += "%04d" % ii
        if idx != len(natoms_list)-1 :
            comb_name += "-"
    path_work = os.path.join(path_sc, comb_name)
    create_path(path_work)
    cwd = os.getcwd()
    to_file = os.path.abspath(to_file)
    os.chdir(path_work)
    try:
        os.symlink(os.path.relpath(to_file), 'POSCAR')
    except FileExistsError:
        pass
    os.chdir(cwd)

def make_combines (dim, natoms) :
    if dim == 1 :
        return [[natoms]]
    else :
        res = []
        for ii in range(natoms+1) :
            rest = natoms - ii
            tmp_combines = make_combines(dim-1, rest)
            for jj in tmp_combines :
                jj.append(ii)
            if len(res) == 0 :
                res = tmp_combines
            else :
                res += tmp_combines
        return res

def place_element (jdata) :
    out_dir = jdata['out_dir']
    super_cell = jdata['super_cell']
    cell_type = class_cell_type(jdata)
    natoms = np.cumprod(super_cell)[-1] * cell_type.numb_atoms()
    elements = jdata['elements']
    path_sc = os.path.join(out_dir, global_dirname_02)
    path_pe = os.path.join(out_dir, global_dirname_02)    
    combines = np.array(make_combines(len(elements), natoms), dtype = int)
    
    assert(os.path.isdir(path_pe))
    cwd = os.getcwd()
    for ii in combines :
        if any(ii == 0) :
            continue
        comb_name = "sys-"
        for idx,jj in enumerate(ii) :            
            comb_name += "%04d" % jj
            if idx != len(ii)-1 :
                comb_name += "-"
        path_pos_in = path_sc
        path_work = os.path.join(path_pe, comb_name)
        create_path(path_work)
        pos_in = os.path.join(path_pos_in, 'POSCAR')
        pos_out = os.path.join(path_work, 'POSCAR')
        poscar_ele(pos_in, pos_out, elements, ii)
        poscar_shuffle(pos_out, pos_out)

def make_vasp_relax (jdata, mdata) :
    out_dir = jdata['out_dir']
    potcars = jdata['potcars']
    cwd = os.getcwd()
    work_dir = os.path.join(out_dir, global_dirname_02)
    assert (os.path.isdir(work_dir))
    work_dir = os.path.abspath(work_dir)

    if os.path.isfile(os.path.join(work_dir, 'INCAR' )) :
        os.remove(os.path.join(work_dir, 'INCAR' ))
    if os.path.isfile(os.path.join(work_dir, 'POTCAR')) :
        os.remove(os.path.join(work_dir, 'POTCAR'))
    shutil.copy2( jdata['relax_incar'], 
                 os.path.join(work_dir, 'INCAR'))
    is_cvasp = False
    if 'cvasp' in mdata['fp_resources'].keys():
        is_cvasp = mdata['fp_resources']['cvasp']
    if is_cvasp:
        cvasp_file=os.path.join(ROOT_PATH,'generator/lib/cvasp.py')
        shutil.copyfile(cvasp_file, os.path.join(work_dir, 'cvasp.py'))
    out_potcar = os.path.join(work_dir, 'POTCAR')
    with open(out_potcar, 'w') as outfile:
        for fname in potcars:
            with open(fname) as infile:
                outfile.write(infile.read())
    
    os.chdir(work_dir)
    
    sys_list = glob.glob('sys-*')
    for ss in sys_list:
        os.chdir(ss)
        ln_src = os.path.relpath(os.path.join(work_dir,'INCAR'))
        try:
           os.symlink(ln_src, 'INCAR')
        except FileExistsError:
           pass
        ln_src = os.path.relpath(os.path.join(work_dir,'POTCAR'))
        try:
           os.symlink(ln_src, 'POTCAR')
        except FileExistsError:
           pass
        os.chdir(work_dir)
    os.chdir(cwd)

def make_scale(jdata):
    out_dir = jdata['out_dir']
    scale = jdata['scale']    
    skip_relax = jdata['skip_relax']    

    cwd = os.getcwd()
    init_path = os.path.join(out_dir, global_dirname_02)
    init_path = os.path.abspath(init_path)
    work_path = os.path.join(out_dir, global_dirname_03)
    os.chdir(init_path)
    init_sys = glob.glob("sys-*")
    init_sys.sort()
    os.chdir(cwd)

    create_path(work_path)
    for ii in init_sys :
        for jj in scale :
            if skip_relax :
                pos_src = os.path.join(os.path.join(init_path, ii), 'POSCAR')
                assert(os.path.isfile(pos_src))
            else :
                try:
                    pos_src = os.path.join(os.path.join(init_path, ii), 'CONTCAR')
                    assert(os.path.isfile(pos_src))
                except:
                    raise RuntimeError("not file %s, vasp relaxation should be run before scale poscar")
            scale_path = os.path.join(work_path, ii)
            scale_path = os.path.join(scale_path, "scale-%.3f" % jj)
            create_path(scale_path)
            os.chdir(scale_path) 
            poscar_scale(pos_src, 'POSCAR', jj)
            os.chdir(cwd)

def pert_scaled(jdata) :
    out_dir = jdata['out_dir']
    scale = jdata['scale']    
    pert_box = jdata['pert_box']
    pert_atom = jdata['pert_atom']
    pert_numb = jdata['pert_numb']
    from_poscar = False 
    if 'from_poscar' in jdata :
        from_poscar = jdata['from_poscar']
    
    cwd = os.getcwd()
    path_sp = os.path.join(out_dir, global_dirname_03)
    assert(os.path.isdir(path_sp))
    os.chdir(path_sp)
    sys_pe = glob.glob('sys-*')
    sys_pe.sort()
    os.chdir(cwd)    

    pert_cmd = os.path.dirname(__file__)
    pert_cmd = os.path.join(pert_cmd, 'tools')
    pert_cmd = os.path.join(pert_cmd, 'create_random_disturb.py')
    pert_cmd = 'python3 ' + pert_cmd + ' -etmax %f -ofmt vasp POSCAR %d %f > /dev/null' %(pert_box, pert_numb, pert_atom)
    for ii in sys_pe :
        for jj in scale :
            path_work = path_sp
            path_work = os.path.join(path_work, ii)
            path_work = os.path.join(path_work, 'scale-%.3f' % jj)
            assert(os.path.isdir(path_work))
            os.chdir(path_work)
            sp.check_call(pert_cmd, shell = True)
            for kk in range(pert_numb) :
                pos_in = 'POSCAR%d.vasp' % (kk+1)
                dir_out = '%06d' % (kk+1)
                create_path(dir_out)
                pos_out = os.path.join(dir_out, 'POSCAR')
                if not from_poscar:
                    poscar_shuffle(pos_in, pos_out)
                else :
                    shutil.copy2(pos_in, pos_out)
                os.remove(pos_in)
            kk = -1
            pos_in = 'POSCAR'
            dir_out = '%06d' % (kk+1)
            create_path(dir_out)
            pos_out = os.path.join(dir_out, 'POSCAR')
            if not from_poscar:
                poscar_shuffle(pos_in, pos_out)
            else :
                shutil.copy2(pos_in, pos_out)
            os.chdir(cwd)

def make_vasp_md(jdata) :
    out_dir = jdata['out_dir']
    potcars = jdata['potcars']
    scale = jdata['scale']   
    pert_numb = jdata['pert_numb'] 
    md_nstep = jdata['md_nstep']

    cwd = os.getcwd()
    path_ps = os.path.join(out_dir, global_dirname_03)
    path_ps = os.path.abspath(path_ps)
    assert(os.path.isdir(path_ps))
    os.chdir(path_ps)
    sys_ps = glob.glob('sys-*')
    sys_ps.sort()
    os.chdir(cwd) 
    path_md = os.path.join(out_dir, global_dirname_04)
    path_md = os.path.abspath(path_md)
    create_path(path_md)
    shutil.copy2(jdata['md_incar'], 
                 os.path.join(path_md, 'INCAR'))
    out_potcar = os.path.join(path_md, 'POTCAR')
    with open(out_potcar, 'w') as outfile:
        for fname in potcars:
            with open(fname) as infile:
                outfile.write(infile.read())
    os.chdir(path_md)
    os.chdir(cwd)    

    for ii in sys_ps :
        for jj in scale :
            for kk in range(pert_numb) :
                path_work = path_md
                path_work = os.path.join(path_work, ii)
                path_work = os.path.join(path_work, "scale-%.3f" % jj)
                path_work = os.path.join(path_work, "%06d" % kk)
                create_path(path_work)
                os.chdir(path_work)                
                path_pos = path_ps
                path_pos = os.path.join(path_pos, ii)
                path_pos = os.path.join(path_pos, "scale-%.3f" % jj)
                path_pos = os.path.join(path_pos, "%06d" % kk)
                init_pos = os.path.join(path_pos, 'POSCAR')
                shutil.copy2 (init_pos, 'POSCAR')
                file_incar = os.path.join(path_md, 'INCAR')
                file_potcar = os.path.join(path_md, 'POTCAR')
                try:
                    os.symlink(os.path.relpath(file_incar), 'INCAR')
                except FileExistsError:
                    pass
                try:
                    os.symlink(os.path.relpath(file_potcar), 'POTCAR')
                except FileExistsError:
                    pass
                 
                os.chdir(cwd)                

def coll_vasp_md(jdata) :
    out_dir = jdata['out_dir']
    md_nstep = jdata['md_nstep']
    scale = jdata['scale']    
    pert_numb = jdata['pert_numb']
    coll_ndata = jdata['coll_ndata']

    cwd = os.getcwd()
    path_md = os.path.join(out_dir, global_dirname_04)
    path_md = os.path.abspath(path_md)
    assert(os.path.isdir(path_md)), "md path should exists"
    os.chdir(path_md)
    sys_md = glob.glob('sys-*')
    sys_md.sort()

    for ii in sys_md :
        os.chdir(ii)
        # convert outcars
        valid_outcars = []
        for jj in scale :
            for kk in range(pert_numb) :
                path_work = os.path.join("scale-%.3f" % jj, "%06d" % kk)
                outcar = os.path.join(path_work, 'OUTCAR')
                #dlog.info("OUTCAR",outcar)
                if os.path.isfile(outcar) :
                    #dlog.info("*"*40)
                    with open(outcar, 'r') as fin:
                        nforce = fin.read().count('TOTAL-FORCE')
                    #dlog.info("nforce is", nforce)
                    #dlog.info("md_nstep", md_nstep)
                    if nforce == md_nstep :
                        valid_outcars.append(outcar)
                    else:
                        dlog.info("WARNING : in directory %s nforce in OUTCAR is not equal to settings in INCAR"%(os.getcwd()))
        arg_cvt = " "
        if len(valid_outcars) == 0:
            raise RuntimeError("MD dir: %s: find no valid outcar in sys %s, "
                               "check if your vasp md simulation is correctly done" 
                               % (path_md, ii)) 

        flag=True
        if ("type_map" in jdata) and isinstance(jdata["type_map"], list):
            type_map = jdata["type_map"]
        else:
            type_map = None 
        for oo in valid_outcars :
            if flag:
                _sys = dpdata.LabeledSystem(oo, type_map= type_map)
                if len(_sys)>0:
                   all_sys=_sys
                   flag=False
                else:
                   pass
            else:
                _sys = dpdata.LabeledSystem(oo, type_map= type_map)
                if len(_sys)>0:
                   all_sys.append(_sys)
        # create deepmd data
        if all_sys.get_nframes() >= coll_ndata :
            all_sys = all_sys.sub_system(np.arange(coll_ndata))
        all_sys.to_deepmd_raw('deepmd')
        all_sys.to_deepmd_npy('deepmd', set_size = all_sys.get_nframes())
        os.chdir(path_md)
    os.chdir(cwd)

def _vasp_check_fin (ii) :
    if os.path.isfile(os.path.join(ii, 'OUTCAR')) :
        with open(os.path.join(ii, 'OUTCAR'), 'r') as fp :
            content = fp.read()
            count = content.count('Elapse')
            if count != 1 :
                return False
    else :
        return False
    return True

def run_vasp_relax(jdata, mdata):
    fp_command = mdata['fp_command']
    fp_group_size = mdata['fp_group_size']
    fp_resources = mdata['fp_resources']
    #machine_type = mdata['fp_machine']['machine_type']
    work_dir = os.path.join(jdata['out_dir'], global_dirname_02)
    
    forward_files = ["POSCAR", "INCAR", "POTCAR"]
    backward_files = ["OUTCAR","CONTCAR"]
    forward_common_files = []
    if 'cvasp' in mdata['fp_resources']:
        if mdata['fp_resources']['cvasp']:
            forward_common_files=['cvasp.py']
    relax_tasks = glob.glob(os.path.join(work_dir, "sys-*"))
    relax_tasks.sort()
    #dlog.info("work_dir",work_dir)
    #dlog.info("relax_tasks",relax_tasks)
    if len(relax_tasks) == 0:
        return

    relax_run_tasks = relax_tasks
    #for ii in relax_tasks : 
    #    if not _vasp_check_fin(ii):
    #        relax_run_tasks.append(ii)
    run_tasks = [os.path.basename(ii) for ii in relax_run_tasks]
    dispatcher = make_dispatcher(mdata['fp_machine'], mdata['fp_resources'], work_dir, run_tasks, fp_group_size)
    #dlog.info(run_tasks)
    dispatcher.run_jobs(fp_resources,
                       [fp_command],
                       work_dir,
                       run_tasks,
                       fp_group_size,
                       forward_common_files,
                       forward_files,
                       backward_files)

def run_vasp_md(jdata, mdata):
    fp_command = mdata['fp_command']
    fp_group_size = mdata['fp_group_size']
    fp_resources = mdata['fp_resources']
    #machine_type = mdata['fp_machine']['machine_type']
    work_dir = os.path.join(jdata['out_dir'], global_dirname_04)
    scale = jdata['scale']   
    pert_numb = jdata['pert_numb'] 
    md_nstep = jdata['md_nstep']

    forward_files = ["POSCAR", "INCAR", "POTCAR"]
    backward_files = ["OUTCAR"]
    forward_common_files = []
    if 'cvasp' in mdata['fp_resources']:
        if mdata['fp_resources']['cvasp']:
            forward_common_files=['cvasp.py']

    path_md = work_dir
    path_md = os.path.abspath(path_md)
    cwd = os.getcwd()
    assert(os.path.isdir(path_md)), "md path should exists"
    md_tasks = glob.glob(os.path.join(work_dir, 'sys-*/scale*/00*'))
    md_tasks.sort()

    if len(md_tasks) == 0:
        return

    md_run_tasks = md_tasks
    #for ii in md_tasks : 
    #    if not _vasp_check_fin(ii):
    #        md_run_tasks.append(ii)

    run_tasks = [ii.replace(work_dir+"/", "") for ii in md_run_tasks]
    #dlog.info("md_work_dir", work_dir)
    #dlog.info("run_tasks",run_tasks)
    dispatcher = make_dispatcher(mdata['fp_machine'], mdata['fp_resources'], work_dir, run_tasks, fp_group_size)
    dispatcher.run_jobs(fp_resources,
                       [fp_command],
                       work_dir,
                       run_tasks,
                       fp_group_size,
                       forward_common_files,
                       forward_files,
                       backward_files)

def gen_init_bulk(args) :
    try:
       import ruamel
       from monty.serialization import loadfn,dumpfn
       warnings.simplefilter('ignore', ruamel.yaml.error.MantissaNoDotYAML1_1Warning)
       jdata=loadfn(args.PARAM)
       if args.MACHINE is not None:
          mdata=loadfn(args.MACHINE)
    except:
       with open (args.PARAM, 'r') as fp :
           jdata = json.load (fp)
       if args.MACHINE is not None:
          with open (args.MACHINE, "r") as fp:
              mdata = json.load(fp)

    if args.MACHINE is not None:
       # Selecting a proper machine
       mdata = decide_fp_machine(mdata)
       #disp = make_dispatcher(mdata["fp_machine"])

    # Decide work path
    out_dir = out_dir_name(jdata)
    jdata['out_dir'] = out_dir
    dlog.info ("# working dir %s" % out_dir)
    # Decide whether to use a given poscar
    from_poscar = False 
    if 'from_poscar' in jdata :
        from_poscar = jdata['from_poscar']
    # Verify md_nstep
    md_nstep_jdata = jdata["md_nstep"]
    try:
        md_incar = jdata['md_incar']
        if os.path.isfile(md_incar):
            standard_incar = incar_upper(Incar.from_file(md_incar))
            nsw_flag = False
            if "NSW" in standard_incar:
                    nsw_flag = True
                    nsw_steps = standard_incar['NSW']
            #dlog.info("nsw_steps is", nsw_steps)
            #dlog.info("md_nstep_jdata is", md_nstep_jdata)
            if nsw_flag:
                if (nsw_steps != md_nstep_jdata):
                    dlog.info("WARNING: your set-up for MD steps in PARAM and md_incar are not consistent!")
                    dlog.info("MD steps in PARAM is %d"%(md_nstep_jdata))
                    dlog.info("MD steps in md_incar is %d"%(nsw_steps))
                    dlog.info("DP-GEN will use settings in md_incar!")
                    jdata['md_nstep'] = nsw_steps
    except KeyError:
        pass
    ## correct element name 
    temp_elements = []
    for ele in jdata['elements']:
        temp_elements.append(ele[0].upper() + ele[1:])
    jdata['elements'] = temp_elements
    dlog.info("Elements are %s"% ' '.join(jdata['elements']))

    ## Iteration 
    stage_list = [int(i) for i in jdata['stages']]
    for stage in stage_list:
        if stage == 1 :
            dlog.info("Current stage is 1, relax")
            create_path(out_dir)
            shutil.copy2(args.PARAM, os.path.join(out_dir, 'param.json'))
            if from_poscar :
                make_super_cell_poscar(jdata)
            else :
                make_unit_cell(jdata)
                make_super_cell(jdata)
                place_element(jdata)
            if args.MACHINE is not None:
               make_vasp_relax(jdata, mdata)
               run_vasp_relax(jdata, mdata)
            else:
               make_vasp_relax(jdata, {"fp_resources":{}})
        elif stage == 2 :
            dlog.info("Current stage is 2, perturb and scale")
            make_scale(jdata)
            pert_scaled(jdata)
        elif stage == 3 :
            dlog.info("Current stage is 3, run a short md")
            make_vasp_md(jdata)
            if args.MACHINE is not None:
               run_vasp_md(jdata, mdata)
        elif stage == 4 :
            dlog.info("Current stage is 4, collect data")
            coll_vasp_md(jdata)
        else :
            raise RuntimeError("unknown stage %d" % stage)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generating initial data for bulk systems.")
    parser.add_argument('PARAM', type=str,
                        help="parameter file, json/yaml format")
    parser.add_argument('MACHINE', type=str,default=None,nargs="?",
                        help="machine file, json/yaml format")
    args = parser.parse_args()
    gen_init_bulk(args)



#!/usr/bin/env python3

import os, glob, argparse, json, warnings, re
import numpy as np
import dpgen.auto_test.lib.lammps as lammps
import dpgen.auto_test.lib.vasp as vasp

global_equi_name = '00.equi'

def comput_e_shift(poscar, task_name) :
    a_types = vasp.get_poscar_types(poscar)
    a_natoms = vasp.get_poscar_natoms(poscar)
    ener_shift = 0
    if len(a_types) > 1 :
        if not os.path.isdir('stables') :
            raise RuntimeError('no dir "stable". Stable energy and volume of components should be computed before calculating formation energy of an alloy')
        for ii in range(len(a_types)) :
            ref_e_file = a_types[ii] + '.' + task_name + '.e'
            ref_e_file = os.path.join('stables', ref_e_file)
            ener = float(open(ref_e_file).read())
            ener_shift += a_natoms[ii] * ener
    return ener_shift

def comput_lmp_nev(conf_dir, task_name,write_shift = False) :
    conf_path = re.sub('confs', global_equi_name, conf_dir)
    conf_path = os.path.abspath(conf_path)
    poscar = os.path.join(conf_path, 'POSCAR')
    ele_types = vasp.get_poscar_types(poscar)

    lmp_path = os.path.join(conf_path, task_name)
    log_lammps = os.path.join(lmp_path, 'log.lammps')
    if os.path.isfile(log_lammps):
        natoms, epa, vpa = lammps.get_nev(log_lammps)
        if write_shift and len(ele_types)>1:
            ener_shift = comput_e_shift(poscar, task_name)
            shift = (epa * natoms - ener_shift) / natoms
            return natoms,epa,vpa,shift
        if len(ele_types)==1:
            stable_dir = 'stables'
            os.makedirs(stable_dir, exist_ok=True)
            name_prefix=os.path.join(stable_dir,'%s.%s' % (ele_types[0], task_name))
            open(name_prefix + '.e', 'w').write('%.16f\n' % (epa))
            open(name_prefix + '.v', 'w').write('%.16f\n' % (vpa))
        return natoms, epa, vpa , None
    else :
        return None, None, None, None

def comput_vasp_nev(jdata, conf_dir, write_shift = False) :

    conf_path = re.sub('confs', global_equi_name, conf_dir)
    conf_path = os.path.abspath(conf_path)
    poscar = os.path.join(conf_path, 'POSCAR')
    ele_types = vasp.get_poscar_types(poscar)

    if 'relax_incar' in jdata.keys():
        vasp_str='vasp-relax_incar'
    else:
        kspacing = jdata['vasp_params']['kspacing']
        vasp_str='vasp-k%.2f' % kspacing

    vasp_path = os.path.join(conf_path, vasp_str)
    outcar = os.path.join(vasp_path, 'OUTCAR')
    # tag_fin = os.path.join(vasp_path, 'tag_finished')
    if not os.path.isfile(outcar) :
        warnings.warn("cannot find OUTCAR in "+vasp_path+" skip")
    elif not vasp.check_finished(outcar):
        warnings.warn("incomplete job "+vasp_path+" use the last frame")
    if os.path.isfile(outcar):
        natoms, epa, vpa = vasp.get_nev(outcar)
        if write_shift and len(ele_types)>1:
            ener_shift = comput_e_shift(poscar, vasp_str)
            shift = (epa * natoms - ener_shift) / natoms
            return natoms,epa,vpa,shift
        if len(ele_types)==1:
            stable_dir = 'stables'
            os.makedirs(stable_dir, exist_ok=True)
            name_prefix=os.path.join(stable_dir,'%s.'% (ele_types[0])+vasp_str)
            open(name_prefix + '.e', 'w').write('%.16f\n' % (epa))
            open(name_prefix + '.v', 'w').write('%.16f\n' % (vpa))
        return natoms, epa, vpa, None
    else :
        return None, None, None, None

def _main():
    parser = argparse.ArgumentParser(
        description="cmpt 00.equi")
    parser.add_argument('TASK', type=str,
                        choices = ['all', 'vasp', 'deepmd', 'meam'],
                        help='the task of generation, vasp or lammps')
    parser.add_argument('PARAM', type=str,
                        help='the json param')
    parser.add_argument('CONF', type=str,
                        help='the dir of conf')
    parser.add_argument('-s','--stable', action = 'store_true',
                        help='the dir of conf')
    args = parser.parse_args()
    with open (args.PARAM, 'r') as fp :
        jdata = json.load (fp)

    if args.TASK == 'all' :
        ln, le, lv = comput_lmp_nev(args.CONF, 'deepmd', args.stable)
        mn, me, mv = comput_lmp_nev(args.CONF, 'meam', args.stable)
        vn, ve, vv = comput_vasp_nev(jdata, args.CONF, args.stable)
        if le == None or ve == None or lv == None or vv == None:
            print("%s" % args.CONF)
        else :
            print("%s\t %8.4f %8.4f %8.4f  %7.3f %7.3f %7.3f  %8.4f %7.3f" % (args.CONF, ve, le, (me), vv, lv, (mv), (le-ve), (lv-vv)))
    elif args.TASK == 'vasp' :
        vn, ve, vv = comput_vasp_nev(jdata, args.CONF, args.stable)
        print("%s\t %8.4f  %7.3f " % (args.CONF, ve, vv))
    elif args.TASK == 'deepmd' :
        ln, le, lv = comput_lmp_nev(args.CONF, 'deepmd', args.stable)
        print("%s\t %8.4f  %7.3f " % (args.CONF, le, lv))
    elif args.TASK == 'meam' :
        ln, le, lv = comput_lmp_nev(args.CONF, 'meam', args.stable)
        print("%s\t %8.4f  %7.3f " % (args.CONF, le, lv))


if __name__ == '__main__' :
    _main()



import sys,os

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..')))

from dpgen.dispatcher.LocalContext import LocalSession
from dpgen.dispatcher.LocalContext import LocalContext
from dpgen.dispatcher.LazyLocalContext import LazyLocalContext
from dpgen.dispatcher.SSHContext import SSHSession
from dpgen.dispatcher.SSHContext import SSHContext
# from dpgen.dispatcher.Dispatcher import FinRecord
from dpgen.dispatcher.Dispatcher import _split_tasks

from dpgen.dispatcher.LocalContext import _identical_files

def setUpModule():
    os.chdir(os.path.abspath(os.path.dirname(__file__)))



#!/usr/bin/env python3

import os,sys,json
import subprocess
from collections import defaultdict

import dpdata

def stat_iter(target_folder, 
            param_file = 'param.json', 
            verbose = True, 
            mute = False):
    jdata={}
    with open(f"{target_folder}/{param_file}") as param_file:
        jdata = json.load(param_file)
    iter_dict = defaultdict(lambda: defaultdict(int))
    output = subprocess.run([f"wc -l {target_folder}/iter.??????/02.fp/*out", ],
        shell=True,stdout=subprocess.PIPE).stdout
    data = output.decode() # split(b'\n')
    for line in data.split('\n'):
        if 'out' in line:
            num, relative_path_doc = line.strip().split(' ')
            path_doc = os.path.abspath(relative_path_doc)
            num = int(num)
            prefix, iter_dirname, stage, out_filename = path_doc.rsplit('/',3) # pylint: disable=unused-variable
            pk_id, out_filename = path_doc.rsplit('/', 1)
            iter = int(iter_dirname.split('.')[-1]) # pylint: disable=unused-variable
            out_id = int(out_filename.strip().split('.')[-2]) # pylint: disable=unused-variable
            out_type = out_filename.strip().split('.')[0]
            iter_dict[pk_id][out_type] += num
    # for ii in 
    output2 = subprocess.run([f"ls -d -1 {target_folder}/iter.??????/02.fp/task.*/OUTCAR", ],
        shell=True,stdout=subprocess.PIPE).stdout
    data2 = output2.decode()
    if verbose:
        # print('find OUTCAR', data2)
        print("use param_jsonfile jdata['type_map']",  jdata['type_map'])
    for line in data2.split('\n'):
        if line:
            # [/home/felix/workplace/SiC/iter.000002/02.fp/task.018.000040/OUTCAR]
            path_doc = os.path.abspath(line)
            pk_id, task_dirname, OUTCAR_filename=path_doc.rsplit('/', 2) # pylint: disable=unused-variable
            try:
                _sys = dpdata.LabeledSystem(path_doc, type_map = jdata['type_map'] )
            except:
                try:
                    _sys = dpdata.LabeledSystem(path_doc.replace('OUTCAR','vasprun.xml'), type_map = jdata['type_map'])
                except:
                    _sys = dpdata.LabeledSystem()
            if len(_sys) == 1:
                pass
            else:
                if verbose:
                    print('OUTCAR not label by dpdata, not convergence or unfinshed', path_doc)
                iter_dict[pk_id]['OUTCAR_not_convergence'] +=1
            iter_dict[pk_id]['OUTCAR_total_count'] +=1
    for pk_id in {**iter_dict}:
        if iter_dict[pk_id]['OUTCAR_total_count']:
            iter_dict[pk_id]['reff']=round(iter_dict[pk_id]['OUTCAR_not_convergence']/iter_dict[pk_id]['OUTCAR_total_count'],5)
    for pk_id, value in iter_dict.items():
        print(f"{pk_id}:candidate:{value['candidate']}"
                f":rest_failed:{value['rest_failed']}"
                f":rest_accurate:{value['rest_accurate']}"
                f":OUTCAR_total_count:{value['OUTCAR_total_count']}"
                f":OUTCAR_not_convergence:{value['OUTCAR_not_convergence']}"
                f":reff:{value['reff']}")






#!/usr/bin/env python3

import os, re, argparse, filecmp, json, glob, sys
import subprocess as sp
import numpy as np
import dpgen.auto_test.lib.vasp as vasp
import dpgen.auto_test.lib.lammps as lammps
from pymatgen.analysis.elasticity.elastic import ElasticTensor
from pymatgen.analysis.elasticity.strain import Strain
from pymatgen.analysis.elasticity.stress import Stress

global_equi_name = '00.equi'
global_task_name = '04.interstitial'

def cmpt_vasp(jdata, conf_dir, supercell, insert_ele) :
    for ii in insert_ele:
        _cmpt_vasp(jdata, conf_dir, supercell, ii)

def _cmpt_vasp(jdata, conf_dir, supercell, insert_ele) :

    if 'relax_incar' in jdata.keys():
        vasp_str='vasp-relax_incar'
    else:
        kspacing = jdata['vasp_params']['kspacing']
        vasp_str='vasp-k%.2f' % kspacing

    equi_path = re.sub('confs', global_equi_name, conf_dir)
    equi_path = os.path.join(equi_path, vasp_str)
    equi_path = os.path.abspath(equi_path)
    equi_outcar = os.path.join(equi_path, 'OUTCAR')
    task_path = re.sub('confs', global_task_name, conf_dir)
    task_path = os.path.join(task_path, vasp_str)
    task_path = os.path.abspath(task_path)

    equi_natoms, equi_epa, equi_vpa = vasp.get_nev(equi_outcar)

    copy_str = "%sx%sx%s" % (supercell[0], supercell[1], supercell[2])
    struct_path_widecard = os.path.join(task_path, 'struct-%s-%s-*' % (insert_ele, copy_str))
    print(struct_path_widecard)
    struct_path_list = glob.glob(struct_path_widecard)
    struct_path_list.sort()
    if len(struct_path_list) == 0:
        print("# cannot find results for conf %s supercell %s" % (conf_dir, supercell))
    sys.stdout.write ("Insert_ele-Struct: Inter_E(eV)  E(eV) equi_E(eV)\n")
    result = os.path.join(task_path,'result')
    with open(result,'w') as fp:
        fp.write('conf_dir:%s\n'% (conf_dir))
        fp.write ("Insert_ele-Struct: Inter_E(eV)  E(eV) equi_E(eV)\n")
        for ii in struct_path_list :
            structure_dir = os.path.basename(ii)
            outcar = os.path.join(ii, 'OUTCAR')
            natoms, epa, vpa = vasp.get_nev(outcar)
            evac = epa * natoms - equi_epa * natoms
            sys.stdout.write ("%s: %7.3f  %7.3f %7.3f \n" % (structure_dir, evac, epa * natoms, equi_epa * natoms))
            fp.write ("%s: %7.3f  %7.3f %7.3f \n" % (structure_dir, evac, epa * natoms, equi_epa * natoms))
    fp.close()

def cmpt_deepmd_reprod_traj(jdata, conf_dir, supercell, insert_ele, task_name) :
    for ii in insert_ele:
        _cmpt_deepmd_reprod_traj(jdata, conf_dir, supercell, ii, task_name)

def _cmpt_deepmd_reprod_traj(jdata, conf_dir, supercell, insert_ele, task_name) :
    if 'relax_incar' in jdata.keys():
        vasp_str='vasp-relax_incar'
    else:
        kspacing = jdata['vasp_params']['kspacing']
        vasp_str='vasp-k%.2f' % kspacing

    conf_path = os.path.abspath(conf_dir)
    task_path = re.sub('confs', global_task_name, conf_path)
    vasp_path = os.path.join(task_path, vasp_str)
    lmps_path = os.path.join(task_path, task_name + vasp_str.replace('vasp',''))
    copy_str = "%sx%sx%s" % (supercell[0], supercell[1], supercell[2])
    struct_widecard = os.path.join(vasp_path, 'struct-%s-%s-*' % (insert_ele,copy_str))
    vasp_struct = glob.glob(struct_widecard)
    vasp_struct.sort()
    cwd=os.getcwd()

    for vs in vasp_struct :
        # compute vasp
        outcar = os.path.join(vs, 'OUTCAR')
        vasp_ener = np.array(vasp.get_energies(outcar))
        vasp_ener_file = os.path.join(vs, 'ener.vasp.out')
        # compute reprod
        struct_basename  = os.path.basename(vs)
        ls = os.path.join(lmps_path, struct_basename)
        frame_widecard = os.path.join(ls, 'frame.*')
        frames = glob.glob(frame_widecard)
        frames.sort()
        lmp_ener = []
        for ii in frames :
            log_lmp = os.path.join(ii, 'log.lammps')
            natoms, epa, vpa = lammps.get_nev(log_lmp)
            lmp_ener.append(epa)
        lmp_ener = np.array(lmp_ener)
        lmp_ener = np.reshape(lmp_ener,  [-1,1])
        lmp_ener_file = os.path.join(ls, 'ener.lmp.out')
        vasp_ener = np.reshape(vasp_ener, [-1,1]) / natoms
        error_start = 1
        lmp_ener -= lmp_ener[-1] - vasp_ener[-1]
        diff = lmp_ener - vasp_ener
        diff = diff[error_start:]
        error = np.linalg.norm(diff) / np.sqrt(np.size(lmp_ener))
        np.savetxt(vasp_ener_file, vasp_ener[error_start:])
        np.savetxt(lmp_ener_file,  lmp_ener[error_start:])
        print(os.path.basename(ls), 'EpA_std_err=',error)

def cmpt_deepmd_lammps(jdata, conf_dir, supercell, insert_ele, task_name) :
    for ii in insert_ele:
        _cmpt_deepmd_lammps(jdata, conf_dir, supercell, ii, task_name)

def _cmpt_deepmd_lammps(jdata, conf_dir, supercell, insert_ele, task_name) :
    equi_path = re.sub('confs', global_equi_name, conf_dir)
    equi_path = os.path.join(equi_path, task_name.split('-')[0])
    equi_path = os.path.abspath(equi_path)
    equi_log = os.path.join(equi_path, 'log.lammps')
    task_path = re.sub('confs', global_task_name, conf_dir)
    task_path = os.path.join(task_path, task_name)
    task_path = os.path.abspath(task_path)

    equi_natoms, equi_epa, equi_vpa = lammps.get_nev(equi_log)

    copy_str = "%sx%sx%s" % (supercell[0], supercell[1], supercell[2])
    struct_path_widecard = os.path.join(task_path, 'struct-%s-%s-*' % (insert_ele,copy_str))
    struct_path_list = glob.glob(struct_path_widecard)
    print(struct_path_widecard)
    struct_path_list.sort()
    if len(struct_path_list) == 0:
        print("# cannot find results for conf %s supercell %s" % (conf_dir, supercell))
    sys.stdout.write ("Insert_ele-Struct: Inter_E(eV)  E(eV) equi_E(eV)\n")
    result = os.path.join(task_path,'result')
    with open(result,'w') as fp:
        fp.write('conf_dir:%s\n'% (conf_dir))
        fp.write ("Insert_ele-Struct: Inter_E(eV)  E(eV) equi_E(eV)\n")
        for ii in struct_path_list :
            structure_dir = os.path.basename(ii)
            lmp_log = os.path.join(ii, 'log.lammps')
            natoms, epa, vpa = lammps.get_nev(lmp_log)
            evac = epa * natoms - equi_epa * natoms
            sys.stdout.write ("%s: %7.3f  %7.3f %7.3f \n" % (structure_dir, evac, epa * natoms, equi_epa * natoms))
            fp.write ("%s: %7.3f  %7.3f %7.3f \n" % (structure_dir, evac, epa * natoms, equi_epa * natoms))
    fp.close()

def _main() :
    parser = argparse.ArgumentParser(
        description="cmpt 04.interstitial")
    parser.add_argument('TASK', type=str,
                        help='the task of generation, vasp or lammps')
    parser.add_argument('PARAM', type=str,
                        help='json parameter file')
    parser.add_argument('CONF', type=str,
                        help='the path to conf')
    parser.add_argument('COPY', type=int, nargs = 3,
                        help='define the supercell')
    parser.add_argument('ELEMENT', type=str, nargs = '+',
                        help='the inserted element')
    args = parser.parse_args()

    with open (args.PARAM, 'r') as fp :
        jdata = json.load (fp)

#    print('# generate %s task with conf %s' % (args.TASK, args.CONF))
    if args.TASK == 'vasp':
        cmpt_vasp(jdata, args.CONF, args.COPY, args.ELEMENT)
    elif args.TASK == 'deepmd' :
        cmpt_deepmd_lammps(jdata, args.CONF, args.COPY, args.ELEMENT, args.TASK)
    elif args.TASK == 'deepmd-reprod' :
        cmpt_deepmd_reprod_traj(jdata, args.CONF, args.COPY, args.ELEMENT, args.TASK)
    elif args.TASK == 'meam' :
        cmpt_deepmd_lammps(jdata, args.CONF, args.COPY, args.ELEMENT, args.TASK)
    elif args.TASK == 'meam-reprod' :
        cmpt_deepmd_reprod_traj(jdata, args.CONF, args.COPY, args.ELEMENT, args.TASK)
    else :
        raise RuntimeError("unknow task ", args.TASK)

if __name__ == '__main__' :
    _main()



import numpy as np

def test_atom_names(testCase, system_1, system_2):
    testCase.assertEqual(system_1.data['atom_names'], 
                         system_2.data['atom_names'])

def test_atom_types(testCase, system_1, system_2):
    testCase.assertEqual(system_1.data['atom_types'][0], 
                         system_2.data['atom_types'][0])
    testCase.assertEqual(system_1.data['atom_types'][1],
                         system_2.data['atom_types'][1])

def test_cell(testCase, system_1, system_2, places = 5):
    testCase.assertEqual(system_1.get_nframes(),
                         system_2.get_nframes())        
    for ff in range(system_1.get_nframes()) :
        for ii in range(3) :
            for jj in range(3) :
                testCase.assertAlmostEqual(system_1.data['cells'][ff][ii][jj], 
                                           system_2.data['cells'][ff][ii][jj], 
                                           places = places,
                                           msg = 'cell[%d][%d][%d] failed' % (ff,ii,jj))

def test_coord(testCase, system_1, system_2, places = 5): 
    testCase.assertEqual(system_1.get_nframes(),
                         system_2.get_nframes())
    # think about direct coord
    tmp_cell = system_1.data['cells']
    tmp_cell = np.reshape(tmp_cell, [-1, 3])
    tmp_cell_norm = np.reshape(np.linalg.norm(tmp_cell, axis = 1), [-1, 3])
    for ff in range(system_1.get_nframes()) :
        for ii in range(sum(system_1.data['atom_numbs'])) :
            for jj in range(3) :
                testCase.assertAlmostEqual(system_1.data['coords'][ff][ii][jj] / tmp_cell_norm[ff][jj], 
                                           system_2.data['coords'][ff][ii][jj] / tmp_cell_norm[ff][jj], 
                                           places = places,
                                           msg = 'coord[%d][%d][%d] failed' % (ff,ii,jj))


class CompSys :
    def test_atom_numbs(self):
        self.assertEqual(self.system_1.data['atom_numbs'], 
                         self.system_2.data['atom_numbs'])

    def test_atom_names(self):
        self.assertEqual(self.system_1.data['atom_names'], 
                         self.system_2.data['atom_names'])

    def test_atom_types(self):
        self.assertEqual(self.system_1.data['atom_types'][0], 
                         self.system_2.data['atom_types'][0])
        self.assertEqual(self.system_1.data['atom_types'][1],
                         self.system_2.data['atom_types'][1])

    def test_orig(self):
        for d0 in range(3) :
            self.assertEqual(self.system_1.data['orig'][d0],
                             self.system_2.data['orig'][d0])

    def test_nframs(self):
        self.assertEqual(self.system_1.get_nframes(),
                         self.system_2.get_nframes())

    def test_cell(self):
        self.assertEqual(self.system_1.get_nframes(),
                         self.system_2.get_nframes())        
        for ff in range(self.system_1.get_nframes()) :
            for ii in range(3) :
                for jj in range(3) :
                    self.assertAlmostEqual(self.system_1.data['cells'][ff][ii][jj], 
                                           self.system_2.data['cells'][ff][ii][jj], 
                                           places = self.places,
                                           msg = 'cell[%d][%d][%d] failed' % (ff,ii,jj))

    def test_coord(self): 
        self.assertEqual(self.system_1.get_nframes(),
                         self.system_2.get_nframes())
        # think about direct coord
        tmp_cell = self.system_1.data['cells']
        tmp_cell = np.reshape(tmp_cell, [-1, 3])
        tmp_cell_norm = np.reshape(np.linalg.norm(tmp_cell, axis = 1), [-1, 3])
        for ff in range(self.system_1.get_nframes()) :
            for ii in range(sum(self.system_1.data['atom_numbs'])) :
                for jj in range(3) :
                    self.assertAlmostEqual(self.system_1.data['coords'][ff][ii][jj] / tmp_cell_norm[ff][jj], 
                                           self.system_2.data['coords'][ff][ii][jj] / tmp_cell_norm[ff][jj], 
                                           places = self.places,
                                           msg = 'coord[%d][%d][%d] failed' % (ff,ii,jj))


class CompLabeledSys (CompSys) :
    def test_energy(self) :
        self.assertEqual(self.system_1.get_nframes(),
                         self.system_2.get_nframes())
        for ff in range(self.system_1.get_nframes()) :
            self.assertAlmostEqual(self.system_1.data['energies'][ff], 
                                   self.system_2.data['energies'][ff], 
                                   places = self.e_places,
                                   msg = 'energies[%d] failed' % (ff))

    def test_force(self) :
        self.assertEqual(self.system_1.get_nframes(),
                         self.system_2.get_nframes())        
        for ff in range(self.system_1.get_nframes()) :
            for ii in range(self.system_1.data['forces'].shape[1]) :
                for jj in range(3) :
                    self.assertAlmostEqual(self.system_1.data['forces'][ff][ii][jj], 
                                           self.system_2.data['forces'][ff][ii][jj], 
                                           places = self.f_places,
                                           msg = 'forces[%d][%d][%d] failed' % (ff,ii,jj))
            
    def test_virial(self) :
        self.assertEqual(self.system_1.get_nframes(),
                         self.system_2.get_nframes())        
        if not self.system_1.has_virial():
            return
        for ff in range(self.system_1.get_nframes()) :
            for ii in range(3) :
                for jj in range(3) :
                    self.assertAlmostEqual(self.system_1['virials'][ff][ii][jj], 
                                           self.system_2['virials'][ff][ii][jj], 
                                           places = self.v_places,
                                           msg = 'virials[%d][%d][%d] failed' % (ff,ii,jj))
            
                    



#!/usr/bin/env python3

import os, re, argparse, filecmp, json, glob, shutil
import subprocess as sp
import numpy as np
import dpgen.auto_test.lib.vasp as vasp
import dpgen.auto_test.lib.lammps as lammps

from dpgen import dlog
from dpgen.generator.lib.vasp import incar_upper
from dpgen import ROOT_PATH
from pymatgen.io.vasp import Incar
from dpgen.generator.lib.vasp import incar_upper

cvasp_file=os.path.join(ROOT_PATH,'generator/lib/cvasp.py')

global_task_name = '00.equi'

'''
link poscar
link potcar
make incar
'''
def make_vasp(jdata, conf_dir) :
    conf_path = os.path.abspath(conf_dir)
    equi_path = re.sub('confs', global_task_name, conf_path)
    os.makedirs(equi_path, exist_ok = True)
    cwd = os.getcwd()
    from_poscar = os.path.join(conf_path, 'POSCAR')
    to_poscar = os.path.join(equi_path, 'POSCAR')
    if os.path.exists(to_poscar) :
        assert(filecmp.cmp(from_poscar, to_poscar))
    else :
        os.chdir(equi_path)
        os.symlink(os.path.relpath(from_poscar), 'POSCAR')
        os.chdir(cwd)
    is_alloy = \
               os.path.exists(
                   os.path.join(
                       os.path.join(conf_path, '..'),
                       'alloy'
                   )
               )    
    # read potcar
    with open(to_poscar,'r') as fp :
        lines = fp.read().split('\n')
        ele_list = lines[5].split()
    potcar_map = jdata['potcar_map']
    potcar_list = []
    for ii in ele_list :
        assert os.path.exists(os.path.abspath(potcar_map[ii])),"No POTCAR in the potcar_map of %s"%(ii)
        potcar_list.append(os.path.abspath(potcar_map[ii]))
        
    # gen incar
    if  'relax_incar' in jdata.keys():
        relax_incar_path = jdata['relax_incar']
        assert(os.path.exists(relax_incar_path))
        relax_incar_path = os.path.abspath(relax_incar_path)
        incar = incar_upper(Incar.from_file(relax_incar_path))
        isif = 3
        if incar.get('ISIF') != isif:
            dlog.info("%s:%s setting ISIF to %d" % (__file__, make_vasp.__name__, isif))
            incar['ISIF'] = isif
        fc = incar.get_string()
        kspacing = incar['KSPACING']
        kgamma = incar['KGAMMA']
        vasp_path = os.path.join(equi_path, 'vasp-relax_incar' )
    else :
        fp_params = jdata['vasp_params']
        ecut = fp_params['ecut']
        ediff = fp_params['ediff']
        npar = fp_params['npar']
        kpar = fp_params['kpar']
        kspacing = fp_params['kspacing']
        kgamma = fp_params['kgamma']
        fc = vasp.make_vasp_relax_incar(ecut, ediff, is_alloy,  True, True, npar, kpar, kspacing, kgamma)
        vasp_path = os.path.join(equi_path, 'vasp-k%.2f' % kspacing)

    os.makedirs(vasp_path, exist_ok = True)
    os.chdir(vasp_path)

    # write incar
    with open('INCAR', 'w') as fp :
        fp.write(fc)

    # gen poscar
    if os.path.exists('POSCAR') :
        os.remove('POSCAR')
    os.symlink(os.path.relpath(to_poscar), 'POSCAR')

    # gen kpoints
    fc = vasp.make_kspacing_kpoints('POSCAR', kspacing, kgamma)
    with open('KPOINTS', 'w') as fp: fp.write(fc)

    #copy cvasp
    if ('cvasp' in jdata) and (jdata['cvasp'] == True):
       shutil.copyfile(cvasp_file, 'cvasp.py')

    # gen potcar
    with open('POTCAR', 'w') as outfile:
        for fname in potcar_list:
            with open(fname) as infile:
                outfile.write(infile.read())
    os.chdir(cwd)

def make_lammps (jdata, conf_dir,task_type) :
    fp_params = jdata['lammps_params']
    model_dir = fp_params['model_dir']
    type_map = fp_params['type_map'] 
    model_dir = os.path.abspath(model_dir)
    model_name =fp_params['model_name']
    deepmd_version = fp_params.get("deepmd_version", "0.12")
    if not model_name and task_type =='deepmd':
        models = glob.glob(os.path.join(model_dir, '*pb'))
        model_name = [os.path.basename(ii) for ii in models]
        assert len(model_name)>0,"No deepmd model in the model_dir"
    else:
        models = [os.path.join(model_dir,ii) for ii in model_name]

    model_param = {'model_name' :      model_name,
                  'param_type':          fp_params['model_param_type'],
                  'deepmd_version' : deepmd_version}
    
    ntypes = len(type_map)
    conf_path = os.path.abspath(conf_dir)
    equi_path = re.sub('confs', global_task_name, conf_path)
    os.makedirs(equi_path, exist_ok = True)
    cwd = os.getcwd()
    from_poscar = os.path.join(conf_path, 'POSCAR')
    to_poscar = os.path.join(equi_path, 'POSCAR')
    if os.path.exists(to_poscar) :
        assert(filecmp.cmp(from_poscar, to_poscar))
    else :
        os.chdir(equi_path)
        os.symlink(os.path.relpath(from_poscar), 'POSCAR')
        os.chdir(cwd)
    # lmp path
    lmp_path = os.path.join(equi_path, task_type)
    os.makedirs(lmp_path, exist_ok = True)    
    print(lmp_path)
    # lmp conf
    conf_file = os.path.join(lmp_path, 'conf.lmp')
    lammps.cvt_lammps_conf(to_poscar, os.path.relpath(conf_file))
    ptypes = vasp.get_poscar_types(to_poscar)
    lammps.apply_type_map(conf_file, type_map, ptypes)    
    # lmp input
    if task_type=='deepmd':
        fc = lammps.make_lammps_equi(os.path.basename(conf_file), 
                                 ntypes, 
                                 lammps.inter_deepmd, 
                                 model_param)
    elif task_type=='meam':
        fc = lammps.make_lammps_equi(os.path.basename(conf_file), 
                                 ntypes, 
                                 lammps.inter_meam, 
                                 model_param)
    with open(os.path.join(lmp_path, 'lammps.in'), 'w') as fp :
        fp.write(fc)
    # link models
    os.chdir(lmp_path)
    for ii in model_name :
        if os.path.exists(ii) :
            os.remove(ii)
    for (ii,jj) in zip(models, model_name) :
        os.symlink(os.path.relpath(ii), jj)
    os.chdir(cwd)


def _main() :
    parser = argparse.ArgumentParser(
        description="gen 00.equi")
    parser.add_argument('TASK', type=str,
                        help='the task of generation, vasp or lammps')
    parser.add_argument('PARAM', type=str,
                        help='json parameter file')
    parser.add_argument('CONF', type=str,
                        help='the path to conf')
    args = parser.parse_args()

    with open (args.PARAM, 'r') as fp :
        jdata = json.load (fp)

#    print('generate %s task with conf %s' % (args.TASK, args.CONF))
    if args.TASK == 'vasp':
        make_vasp(jdata, args.CONF)               
    elif args.TASK == 'deepmd' or args.TASK =='meam' :
        make_lammps(jdata, args.CONF,args.TASK)
    else :
        raise RuntimeError("unknow task ", args.TASK)
    
if __name__ == '__main__' :
    _main()




import dpdata
import numpy as np

default_config={
  "GLOBAL": {
    "PROJECT": "DPGEN"
  },
  "FORCE_EVAL": {
    "METHOD": "QS",
    "STRESS_TENSOR": "ANALYTICAL",
    "DFT": {
      "BASIS_SET_FILE_NAME": "./cp2k_basis_pp_file/BASIS_MOLOPT",
      "POTENTIAL_FILE_NAME": "./cp2k_basis_pp_file/GTH_POTENTIALS",
      "CHARGE": 0,
      "UKS": "F",
      "MULTIPLICITY": 1,
      "MGRID": {
        "CUTOFF": 400,
        "REL_CUTOFF": 50,
        "NGRIDS": 4
      },
      "QS": {
        "EPS_DEFAULT": "1.0E-12"
      },
      "SCF": {
        "SCF_GUESS": "ATOMIC",
        "EPS_SCF": "1.0E-6",
        "MAX_SCF": 50,
        "OT": {
          "MINIMIZER": "DIIS",
          "PRECONDITIONER": "FULL_SINGLE_INVERSE"
        }
      },
      "XC": {
        "XC_FUNCTIONAL": {
          "_": "PBE"
        }

      }
    },
    "SUBSYS": {
        "CELL":{
            "A": "10 .0 .0",
            "B": ".0 10 .0",
            "C": ".0 .0 10"
            },
      "COORD": {"@include": "coord.xyz"},
      "KIND": {
          "_": ["H","C","N"],
          "POTENTIAL": ["GTH-PBE-q1","GTH-PBE-q4", "GTH-PBE-q5"],
          "BASIS_SET": ["DZVP-MOLOPT-GTH","DZVP-MOLOPT-GTH","DZVP-MOLOPT-GTH"]
          }
      },
    "PRINT": {
      "FORCES": {
        "_": "ON"
    }
  }
}
}

def update_dict(old_d, update_d):
    """
    a method to recursive update dict
    :old_d: old dictionary
    :update_d: some update value written in dictionary form
    """
    import collections
    for k, v in update_d.items():
        if (k in old_d and isinstance(old_d[k], dict) and isinstance(update_d[k], collections.Mapping)):
            update_dict(old_d[k], update_d[k])
        else:
            old_d[k] = update_d[k]

def iterdict(d, out_list, flag=None):
    """
    :doc: a recursive expansion of dictionary into cp2k input
    :k: current key
    :v: current value
    :d: current dictionary under expansion
    :flag: used to record dictionary state. if flag is None,
    it means we are in top level dict. flag is a string.
    """
    for k,v in d.items():
        k=str(k) # cast key into string
        #if value is dictionary
        if isinstance(v, dict):
            # flag == None, it is now in top level section of cp2k
            if flag==None :
                out_list.append("&"+k)
                out_list.append("&END "+k)
                iterdict(v, out_list, k)
            # flag is not None, now it has name of section
            else:
                index = out_list.index("&END " + flag)
                out_list.insert(index, "&"+k)
                out_list.insert(index+1,"&END "+k )
                iterdict(v, out_list, k)
        elif isinstance(v, list):
#            print("we have encountered the repeat section!")
            index = out_list.index("&"+flag)
            # delete the current constructed repeat section
            del out_list[index:index+2]
            # do a loop over key and corresponding list
            k_tmp_list = []
            v_list_tmp_list = []
            for k_tmp, v_tmp in d.items():
                k_tmp_list.append(str(k_tmp))
                v_list_tmp_list.append(v_tmp)
            for repeat_keyword in zip(*v_list_tmp_list):
                out_list.insert(index,"&" + flag)
                out_list.insert(index+1, "&END " + flag)
                for idx, k_tmp in enumerate(k_tmp_list):
                    if k_tmp == "_":
                        out_list[index] = "&" + flag + " " + repeat_keyword[idx]
                    else:
                        out_list.insert(index+1, k_tmp+" "+repeat_keyword[idx])

            break

        else:
            v=str(v)
            if flag==None :
                out_list.append(k+" "+v)
                print (k,":",v)
            else:
                if k == "_":
                    index = out_list.index("&" + flag)
                    out_list[index] = ("&" + flag + " " + v)

                else:
                    index = out_list.index("&END "+flag)
                    out_list.insert(index, k+" "+v)


def make_cp2k_input(sys_data, fp_params):
    #covert cell to cell string
    cell = sys_data['cells'][0]
    cell = np.reshape(cell, [3,3])
    cell_a = np.array2string(cell[0,:])
    cell_a = cell_a[1:-1]
    cell_b = np.array2string(cell[1,:])
    cell_b = cell_b[1:-1]
    cell_c = np.array2string(cell[2,:])
    cell_c = cell_c[1:-1]

    #get update from user
    user_config=fp_params
    #get update from cell
    cell_config={"FORCE_EVAL":{
        "SUBSYS":{
            "CELL":{
                "A": cell_a,
                "B": cell_b,
                "C": cell_c
                }
            }
        }
            }

    update_dict(default_config, user_config)
    update_dict(default_config, cell_config)
    #output list
    input_str = []
    iterdict(default_config, input_str)
    string="\n".join(input_str)
    return string




def make_cp2k_xyz(sys_data):
    #get structral information
    atom_names = sys_data['atom_names']
    atom_types = sys_data['atom_types']

    #write coordinate to xyz file used by cp2k input
    coord_list = sys_data['coords'][0]
    u = np.array(atom_names)
    atom_list = u[atom_types]
    x = '\n'
    for kind, coord in zip(atom_list, coord_list) :
        x += str(kind) + ' ' + str(coord[:])[1:-1] + '\n'
    return x







#!/usr/bin/env python3

import numpy as np
import dpdata
import ase,os


def make_one(out_dir) :
    # [0.5, 1)
    [aa,bb,cc] = np.random.random(3) * 0.5 + 0.5
    # [1, 179)
    [alpha,beta,gamma] = np.random.random(3) * (178 / 180) + 1
    # make cell
    cell = ase.geometry.cellpar_to_cell([aa,bb,cc,alpha,beta,gamma])
    sys = dpdata.System('POSCAR')
    sys['cells'][0] = cell
    os.makedirs(out_dir, exist_ok=True)
    sys.to_vasp_poscar(os.path.join(out_dir, 'POSCAR'))
    

ntest = 30
for ii in range(ntest) :
     out_dir = 'test.%03d' % ii
     make_one(out_dir)



#!/usr/bin/python3 

import os
import numpy as np
from pymatgen.io.vasp import Incar

def _make_vasp_incar_dict (ecut, ediff, npar, kpar, 
                           kspacing = 0.5, kgamma = True, 
                           smearing = None, sigma = None, 
                           metagga = None) :
    incar_dict = {}
    incar_dict['PREC'] = 'A'
    incar_dict['ENCUT'] = ecut
    incar_dict['ISYM'] = 0
    incar_dict['ALGO'] = 'fast'
    incar_dict['EDIFF'] = ediff
    incar_dict['LREAL'] = 'A'
    incar_dict['NPAR'] = npar
    incar_dict['KPAR'] = kpar
    incar_dict['NELMIN'] = 4
    incar_dict['ISIF'] = 2
    if smearing is not None :
        incar_dict['ISMEAR'] = smearing        
    if sigma is not None :
        incar_dict['SIGMA'] = sigma
    incar_dict['IBRION'] = -1
    incar_dict['NSW'] = 0
    incar_dict['LWAVE'] = 'F'
    incar_dict['LCHARG'] = 'F'
    incar_dict['PSTRESS'] = 0
    incar_dict['KSPACING'] = kspacing
    if kgamma:
        incar_dict['KGAMMA'] = 'T'
    else :
        incar_dict['KGAMMA'] = 'F'
    if metagga is not None :
        incar_dict['LASPH'] = 'T'
        incar_dict['METAGGA'] = metagga
    return incar_dict

def _update_incar_dict(incar_dict_, user_dict) :
    if user_dict is None:
        return incar_dict_
    incar_dict = incar_dict_
    for ii in user_dict :
        ci = ii.upper()
        incar_dict[ci] = user_dict[ii]
    return incar_dict

def write_incar_dict(incar_dict) :
    lines = []
    for key in incar_dict:
        if (type(incar_dict[key]) == bool):
            if incar_dict[key]:
                rs = 'T'
            else :
                rs = 'F'
        else :
            rs = str(incar_dict[key])
        lines.append('%s=%s' % (key, rs))
    return '\n'.join(lines)


def _make_smearing(fp_params) :
    smearing = None
    sigma = None
    if 'smearing' in fp_params :
        smearing = fp_params['smearing']
    if 'sigma' in fp_params :
        sigma = fp_params['sigma']
    if smearing == None :
        return None, sigma
    smearing_method = (smearing.split(':')[0]).lower()
    if smearing_method == 'mp' :
        order = 1
        if len(smearing.split(':')) == 2 :
            order = int(smearing.split(':')[1])
        return order, sigma
    elif smearing_method == 'gauss' :
        return 0, sigma
    elif smearing_method == 'fd' :
        return -1, sigma
    else :
        raise RuntimeError("unsuppported smearing method %s " % smearing_method)

def _make_metagga(fp_params) :
    metagga = None
    if 'metagga' in fp_params :
        metagga = fp_params['metagga']
    if metagga == 'NONE':
        metagga = None
    elif metagga not in [None,'SCAN', 'TPSS', 'RTPSS', 'M06L', 'MBJ'] :
        raise RuntimeError ("unknown metagga method " + metagga) 
    return metagga
    
def make_vasp_incar_user_dict(fp_params) :
    ecut = fp_params['ecut']
    ediff = fp_params['ediff']
    npar = fp_params['npar']
    kpar = fp_params['kpar']
    kspacing = fp_params['kspacing']
    if 'user_vasp_params' in fp_params :
        user_dict = fp_params['user_vasp_params']
    else :
        user_dict = None
    smearing, sigma = _make_smearing(fp_params)
    metagga = _make_metagga(fp_params)
    incar_dict = _make_vasp_incar_dict(ecut, ediff, npar, kpar, 
                                       kspacing = kspacing, kgamma = False, 
                                       smearing = smearing, sigma = sigma, 
                                       metagga = metagga
    )
    incar_dict = _update_incar_dict(incar_dict, user_dict)
    incar = write_incar_dict(incar_dict)
    return incar
    
def incar_upper(dincar):
    standard_incar={}
    for key,val in dincar.items():
        standard_incar[key.upper()]=val
    return Incar(standard_incar)



import os,sys,json,glob,shutil,copy
import dpdata
import numpy as np
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'generator'
from .context import make_model_devi
from .context import parse_cur_job
from .context import parse_cur_job_revmat
from .context import param_file
from .context import machine_file
from .context import my_file_cmp
from .context import setUpModule
from .context import find_only_one_key
from .context import revise_lmp_input_model
from .context import revise_lmp_input_dump
from .context import revise_lmp_input_plm
from .context import revise_by_keys
from .comp_sys import test_atom_names
from .comp_sys import test_atom_types
from .comp_sys import test_coord
from .comp_sys import test_cell

def _make_fake_models(idx, numb_models) :
    train_dir = os.path.join('iter.%06d' % idx, 
                             '00.train')
    os.makedirs(train_dir, exist_ok = True)
    pwd = os.getcwd()
    os.chdir(train_dir)
    for ii in range(numb_models) :
        os.makedirs('%03d' % ii, exist_ok = True)
        with open(os.path.join('%03d' % ii, 'forzen_model.pb'), 'w') as fp:
            fp.write(str(ii))
        if not os.path.isfile('graph.%03d.pb' % ii) :
            os.symlink(os.path.join('%03d' % ii, 'forzen_model.pb'), 
                       'graph.%03d.pb' % ii)
    os.chdir(pwd)


def _check_confs(testCase, idx, jdata) :
    md_dir = os.path.join('iter.%06d' % idx, 
                          '01.model_devi')
    tasks = glob.glob(os.path.join(md_dir, 'task.*'))
    tasks.sort()
    cur_job = jdata['model_devi_jobs'][idx]
    sys_idx = cur_job['sys_idx']
    sys_configs = jdata['sys_configs']
    poscars = []
    for ii in sys_idx :
        sys_poscars = []
        for ss in sys_configs[ii]:
            tmp_poscars = glob.glob(ss)
            sys_poscars += tmp_poscars
        sys_poscars.sort()
        poscars.append(sys_poscars)
    for ii in tasks :
        conf_file = os.path.join(ii, 'conf.lmp')
        l_conf_file = os.path.basename(os.readlink(conf_file))
        poscar_file = poscars[int(l_conf_file.split('.')[0])][int(l_conf_file.split('.')[1])]
        sys_0 = dpdata.System(conf_file, type_map = jdata['type_map'])
        sys_1 = dpdata.System(poscar_file, type_map = jdata['type_map'])
        test_atom_names(testCase, sys_0, sys_1)
        test_atom_types(testCase, sys_0, sys_1)
        test_cell(testCase, sys_0, sys_1)
        test_coord(testCase, sys_0, sys_1)
        

def _check_pb(testCase, idx) :
    md_dir = os.path.join('iter.%06d' % idx, 
                          '01.model_devi')
    tr_dir = os.path.join('iter.%06d' % idx, 
                          '00.train')
    md_pb = glob.glob(os.path.join(md_dir, 'grapb*pb'))
    tr_pb = glob.glob(os.path.join(tr_dir, 'grapb*pb'))
    md_pb.sort()
    tr_pb.sort()
    for ii,jj in zip(md_pb, tr_pb) :
        my_file_cmp(testCase,ii,jj)

        
def _check_traj_dir(testCase, idx) :
    md_dir = os.path.join('iter.%06d' % idx, 
                          '01.model_devi')
    tasks = glob.glob(os.path.join(md_dir, 'task.*'))
    tasks.sort()
    for ii in tasks:
        testCase.assertTrue(os.path.isdir(os.path.join(ii, 'traj')))


def _get_lammps_pt(lmp_input) :
    with open(lmp_input) as fp: 
        for ii in fp:
            if 'variable' in ii and 'TEMP' in ii :
                lt = float(ii.split()[3])
            if 'variable' in ii and 'PRES' in ii :
                lp = float(ii.split()[3])
    return lt,lp

def _check_pt(testCase, idx, jdata) :
    md_dir = os.path.join('iter.%06d' % idx, 
                          '01.model_devi')
    tasks = glob.glob(os.path.join(md_dir, 'task.*'))
    tasks.sort()
    cur_job = jdata['model_devi_jobs'][idx]
    ensemble, nsteps, trj_freq, temps, press, pka_e, dt = parse_cur_job(cur_job)
    testCase.assertTrue(ensemble, 'npt')
    # get poscars
    sys_idx = cur_job['sys_idx']
    sys_configs = jdata['sys_configs']
    poscars = []
    for ii in sys_idx :
        sys_poscars = []
        for ss in sys_configs[ii]:
            tmp_poscars = glob.glob(ss)
            sys_poscars += tmp_poscars
        sys_poscars.sort()
        poscars.append(sys_poscars)
    for sidx,ii in enumerate(poscars) :
        count = 0
        for ss in ii:
            for tt in temps:
                for pp in press:
                    task_dir = os.path.join(md_dir, 'task.%03d.%06d' % (sidx, count))
                    lt, lp = _get_lammps_pt(os.path.join(task_dir, 'input.lammps'))
                    testCase.assertAlmostEqual(lt, tt)
                    testCase.assertAlmostEqual(lp, pp)
                    count += 1
                    

class TestMakeModelDevi(unittest.TestCase):
    def tearDown(self):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')

    def test_make_model_devi (self) :        
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        _make_fake_models(0, jdata['numb_models'])
        make_model_devi(0, jdata, mdata)
        _check_pb(self, 0)
        _check_confs(self, 0, jdata)
        _check_traj_dir(self, 0)
        _check_pt(self, 0, jdata)
        #shutil.rmtree('iter.000000')        

    def test_make_model_devi_nopbc_npt (self) :        
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_file, 'r') as fp :
            jdata = json.load (fp)
            jdata['model_devi_nopbc'] = True
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        _make_fake_models(0, jdata['numb_models'])
        cwd = os.getcwd()
        with self.assertRaises(RuntimeError) :
            make_model_devi(0, jdata, mdata)        
        os.chdir(cwd)

    def test_make_model_devi_nopbc_nvt (self) :        
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        with open (param_file, 'r') as fp :
            jdata = json.load (fp)
            jdata['model_devi_nopbc'] = True
            jdata['model_devi_jobs'][0]['ensemble'] = 'nvt'
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        _make_fake_models(0, jdata['numb_models'])
        make_model_devi(0, jdata, mdata)
        _check_pb(self, 0)
        # _check_confs(self, 0, jdata)
        _check_traj_dir(self, 0)
        _check_pt(self, 0, jdata)
        # shutil.rmtree('iter.000000')


class TestMakeModelDeviRevMat(unittest.TestCase):
    def tearDown(self):
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')

    def test_make_model_devi (self) :        
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        jdata = {
            "type_map":		["Mg", "Al"],
            "mass_map":		[24, 27],
            "init_data_prefix":	"data",
            "init_data_sys":	["deepmd"],
            "init_batch_size":	[16],            
            "sys_configs_prefix": os.getcwd(),
            "sys_configs":	[
	        ["data/al.fcc.02x02x02/01.scale_pert/sys-0032/scale*/000001/POSCAR"],
	        ["data/al.fcc.02x02x02/01.scale_pert/sys-0032/scale*/000000/POSCAR"]
            ],
            "numb_models":	4,
            "shuffle_poscar":	False,
            "model_devi_f_trust_lo":	0.050,
            "model_devi_f_trust_hi":	0.150,
            "model_devi_e_trust_lo":	1e10,
            "model_devi_e_trust_hi":	1e10,
            "model_devi_plumed":        True,
            "model_devi_jobs":	[
                {"sys_idx": [0, 1], 'traj_freq': 10,  "template":{"lmp": "lmp/input.lammps", "plm": "lmp/input.plumed"},
                 "rev_mat":{
                     "lmp": {"V_NSTEPS": [1000], "V_TEMP": [50, 100], "V_PRES": [1, 10]}, "plm": {"V_DIST0":  [3,4], "V_DIST1": [5, 6]}
                 }}
            ]
        }
        mdata = {'deepmd_version': '1'}
        _make_fake_models(0, jdata['numb_models'])
        make_model_devi(0, jdata, mdata)
        _check_pb(self, 0)
        _check_confs(self, 0, jdata)
        _check_traj_dir(self, 0)
        # check the first task
        md_dir = os.path.join('iter.%06d' % 0, '01.model_devi')
        tasks = glob.glob(os.path.join(md_dir, 'task.*'))
        tasks.sort()
        # 4 accounts for 2 systems each with 2 frames
        self.assertEqual(len(tasks), (len(jdata['model_devi_jobs'][0]['rev_mat']['lmp']['V_NSTEPS']) * 
                                      len(jdata['model_devi_jobs'][0]['rev_mat']['lmp']['V_TEMP']) * 
                                      len(jdata['model_devi_jobs'][0]['rev_mat']['lmp']['V_PRES']) * 
                                      len(jdata['model_devi_jobs'][0]['rev_mat']['plm']['V_DIST0']) * 
                                      len(jdata['model_devi_jobs'][0]['rev_mat']['plm']['V_DIST1']) * 
                                      4))

        cur_job = jdata['model_devi_jobs'][0]
        rev_keys = ['V_NSTEPS', 'V_TEMP', 'V_PRES', 'V_DIST0', 'V_DIST1']
        rev_matrix = []
        for i0 in cur_job['rev_mat']['lmp']['V_NSTEPS']:
            for i1 in cur_job['rev_mat']['lmp']['V_TEMP']:
                for i2 in cur_job['rev_mat']['lmp']['V_PRES']:
                    for i3 in cur_job['rev_mat']['plm']['V_DIST0']:
                        for i4 in cur_job['rev_mat']['plm']['V_DIST1']:
                            rev_matrix.append([i0, i1, i2, i3, i4])        
        numb_rev = len(rev_matrix)
        for ii in range(len(tasks)):
            with open(os.path.join(tasks[ii], 'job.json')) as fp:
                rev_values = rev_matrix[ii % numb_rev]
                job_recd = json.load(fp)
                for kk in job_recd.keys():
                    kidx = rev_keys.index(kk)
                self.assertEqual(rev_values[kidx], job_recd[kk])

        cwd_ = os.getcwd()
        os.chdir(tasks[0])
        with open('input.lammps') as fp:
            lines = fp.readlines()
            for ii in lines:
                if 'variable' in ii and 'TEMP' in ii:
                    self.assertEqual('variable TEMP equal 50', 
                                     ' '.join(ii.split()))
                if 'variable' in ii and 'PRES' in ii:
                    self.assertEqual('variable PRES equal 1', 
                                     ' '.join(ii.split()))
                if 'variable' in ii and 'NSTEPS' in ii:
                    self.assertEqual('variable NSTEPS equal 1000', 
                                     ' '.join(ii.split()))
        with open('input.plumed') as fp:
            lines = fp.readlines()
            for ii in lines:
                if 'RESTRAINT' in ii:
                    self.assertEqual('RESTRAINT ARG=d1,d2 AT=3,5 KAPPA=150.0,150.0 LABEL=restraint', 
                                     ' '.join(ii.split()))
        os.chdir(cwd_)


    def test_make_model_devi_null (self) :        
        if os.path.isdir('iter.000000') :
            shutil.rmtree('iter.000000')
        jdata = {
            "type_map":		["Mg", "Al"],
            "mass_map":		[24, 27],
            "init_data_prefix":	"data",
            "init_data_sys":	["deepmd"],
            "init_batch_size":	[16],            
            "sys_configs_prefix": os.getcwd(),
            "sys_configs":	[
	        ["data/al.fcc.02x02x02/01.scale_pert/sys-0032/scale*/000001/POSCAR"],
	        ["data/al.fcc.02x02x02/01.scale_pert/sys-0032/scale*/000000/POSCAR"]
            ],
            "numb_models":	4,
            "shuffle_poscar":	False,
            "model_devi_f_trust_lo":	0.050,
            "model_devi_f_trust_hi":	0.150,
            "model_devi_e_trust_lo":	1e10,
            "model_devi_e_trust_hi":	1e10,
            "model_devi_plumed":        True,
            "model_devi_jobs":	[
                {"sys_idx": [0, 1], 'traj_freq': 10,  "template":{"lmp": "lmp/input.lammps", "plm": "lmp/input.plumed"},
                }
            ]
        }
        mdata = {'deepmd_version': '1'}
        _make_fake_models(0, jdata['numb_models'])
        make_model_devi(0, jdata, mdata)
        _check_pb(self, 0)
        _check_confs(self, 0, jdata)
        _check_traj_dir(self, 0)
        # check the first task
        md_dir = os.path.join('iter.%06d' % 0, '01.model_devi')
        tasks = glob.glob(os.path.join(md_dir, 'task.*'))
        # 4 accounts for 2 systems each with 2 frames
        self.assertEqual(len(tasks), (4))
        tasks.sort()
        cwd_ = os.getcwd()
        os.chdir(tasks[0])
        with open('input.lammps') as fp:
            lines = fp.readlines()
            for ii in lines:
                if 'variable' in ii and 'TEMP' in ii:
                    self.assertEqual('variable TEMP equal V_TEMP', 
                                     ' '.join(ii.split()))
                if 'variable' in ii and 'PRES' in ii:
                    self.assertEqual('variable PRES equal V_PRES', 
                                     ' '.join(ii.split()))
                if 'variable' in ii and 'NSTEPS' in ii:
                    self.assertEqual('variable NSTEPS equal V_NSTEPS', 
                                     ' '.join(ii.split()))
        with open('input.plumed') as fp:
            lines = fp.readlines()
            for ii in lines:
                if 'RESTRAINT' in ii:
                    self.assertEqual('RESTRAINT ARG=d1,d2 AT=V_DIST0,V_DIST1 KAPPA=150.0,150.0 LABEL=restraint', 
                                     ' '.join(ii.split()))
        os.chdir(cwd_)




class TestParseCurJobRevMat(unittest.TestCase):
    def setUp(self):
        self.cur_job = {
            "sys_idx": [0, 1],
            "template":{"lmp": "lmp/input.lammps", "plm": "lmp/input.plumed"},
            "rev_mat":{
                "lmp": {"V_NSTEPS": [1000], "V_TEMP": [50, 100], "V_PRES": [1, 10]}, "plm": {"V_DIST0":  [3,4], "V_DIST1": [5, 6]}
            }
        }
        self.ref_matrix = []
        for i0 in self.cur_job['rev_mat']['lmp']['V_NSTEPS']:
            for i1 in self.cur_job['rev_mat']['lmp']['V_TEMP']:
                for i2 in self.cur_job['rev_mat']['lmp']['V_PRES']:
                    for i3 in self.cur_job['rev_mat']['plm']['V_DIST0']:
                        for i4 in self.cur_job['rev_mat']['plm']['V_DIST1']:
                            self.ref_matrix.append([i0, i1, i2, i3, i4])
        self.ref_keys = ['V_NSTEPS', 'V_TEMP', 'V_PRES', 'V_DIST0', 'V_DIST1']
        self.ref_nlmp = 3

    def test_parse_cur_job(self):
        rk, rm, nl = parse_cur_job_revmat(self.cur_job, use_plm = True)
        self.assertEqual(rk, self.ref_keys)
        self.assertEqual(nl, self.ref_nlmp)
        self.assertEqual(rm, self.ref_matrix)
        

class MakeModelDeviByReviseMatrix(unittest.TestCase):
    def test_find_only_one_key_1(self):
        lines = ['aaa bbb ccc\n', 'bbb ccc\n', 'ccc bbb ccc\n']
        idx = find_only_one_key(lines, ['bbb', 'ccc'])
        self.assertEqual(idx, 1)

    def test_find_only_one_key_0(self):        
        lines = ['aaa bbb\n', 'bbb aaa\n', 'ccc ddd\n']
        with self.assertRaises(RuntimeError):
            idx = find_only_one_key(lines, ['ccc','eee'])

    def test_find_only_one_key_2(self):        
        lines = ['aaa bbb\n', 'bbb ccc\n', 'bbb ccc\n', 'fff eee\n']
        with self.assertRaises(RuntimeError):
            idx = find_only_one_key(lines, ['bbb','ccc'])

    def test_revise_lmp_input_model_0(self):
        lines = ['foo\n', 'pair_style deepmd  aaa ccc fff\n', 'bar\n', '\n']
        ref_lines = copy.deepcopy(lines)
        lines = revise_lmp_input_model(lines, ['model0', 'model1'], 10, '0.1')
        for ii in [0, 2, 3] :
            self.assertEqual(lines[ii], ref_lines[ii])
        tmp = " ".join(lines[1].split())
        self.assertEqual(tmp, "pair_style deepmd model0 model1 10 model_devi.out")
        
    def test_revise_lmp_input_model_1(self):
        lines = ['foo\n', 'pair_style deepmd aaa ccc fff\n', 'bar\n', '\n']
        ref_lines = copy.deepcopy(lines)
        lines = revise_lmp_input_model(lines, ['model0', 'model1'], 10, '1')
        for ii in [0, 2, 3] :
            self.assertEqual(lines[ii], ref_lines[ii])
        tmp = " ".join(lines[1].split())
        self.assertEqual(tmp, "pair_style deepmd model0 model1 out_freq 10 out_file model_devi.out")
        
    def test_revise_lmp_input_dump(self):
        lines = ['foo\n', 'dump dpgen_dump ccc fff\n', 'bar\n', '\n']
        ref_lines = copy.deepcopy(lines)
        lines = revise_lmp_input_dump(lines, 10)
        for ii in [0, 2, 3] :
            self.assertEqual(lines[ii], ref_lines[ii])
        tmp = " ".join(lines[1].split())
        self.assertEqual(tmp, "dump dpgen_dump all custom 10 traj/*.lammpstrj id type x y z")
        
    def test_revise_lmp_input_plm(self):
        lines = ['foo\n', 'fix dpgen_plm ccc fff\n', 'bar\n', '\n']
        ref_lines = copy.deepcopy(lines)
        lines = revise_lmp_input_plm(lines, 'input.plumed')
        for ii in [0, 2, 3] :
            self.assertEqual(lines[ii], ref_lines[ii])
        tmp = " ".join(lines[1].split())
        self.assertEqual(tmp, "fix dpgen_plm all plumed plumedfile input.plumed outfile output.plumed")
        
    def test_revise_by_key(self):
        lines = ['foo\n', 'aaa\n', 'bar\n', 'bbb\n', '\n']
        ref_lines = copy.deepcopy(lines)
        lines = revise_by_keys(lines, ['aaa', 'bbb'], ['ccc','ddd'])
        for ii in [0, 2, 4] :
            self.assertEqual(lines[ii], ref_lines[ii])
        tmp = " ".join(lines[1].split())
        self.assertEqual(tmp, "ccc")
        tmp = " ".join(lines[3].split())
        self.assertEqual(tmp, "ddd")
        

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import os,sys,json,glob,shutil,dpdata
import numpy as np
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
__package__ = 'generator'
from .context import make_train
from .context import param_file
from .context import param_file_v1
from .context import param_file_v1_et
from .context import machine_file
from .context import machine_file_v1
from .context import setUpModule

def _comp_sys_files (sys0, sys1) :
    pwd = os.getcwd()
    os.chdir(sys0) 
    files = glob.glob('*.raw') 
    set_files = glob.glob('set.*/*npy') 
    # files += set_files
    os.chdir(pwd)
    for ii in files :
        with open(os.path.join(sys0, ii)) as fp0 :
            with open(os.path.join(sys1, ii)) as fp1:
                if fp0.read() != fp1.read() :
                    return False
    for ii in set_files:
        t0 = np.load(os.path.join(sys0, ii))
        t1 = np.load(os.path.join(sys1, ii))
        if np.linalg.norm(t0-t1) > 1e-12 :
            return False
    return True

def _comp_init_data(testCase, iter_idx, init_data_prefix, init_data_sys) :
    for ii in init_data_sys :
        sys0 = os.path.join(init_data_prefix, ii) 
        sys1 = os.path.join('iter.%06d' % iter_idx, 
                            '00.train',
                            'data.init', 
                            ii)
        testCase.assertTrue(_comp_sys_files(sys0, sys1),
                            'systems %s %s are not identical' % (sys0, sys1))

def _check_numb_models(testCase, iter_idx, numb_models) :
    models = glob.glob(os.path.join('iter.%06d' % iter_idx, 
                                    '00.train', 
                                    '[0-9][0-9][0-9]'))
    testCase.assertTrue(len(models), numb_models)


def _check_model_inputs(testCase, iter_idx, jdata) :
    train_param = jdata.get('train_param', 'input.json')
    numb_models = jdata['numb_models']
    default_training_param = jdata['default_training_param']
    init_data_sys = [os.path.join('..', 'data.init', ii) for ii in jdata['init_data_sys']]
    init_batch_size = jdata['init_batch_size']
    sys_batch_size = jdata['sys_batch_size']
    if iter_idx > 0 :
        systems = glob.glob(os.path.join('iter.*', '02.fp', 'data.*'))
        for ii in systems :
            init_data_sys.append(os.path.join('..', 'data.iters', ii))
            sys_idx = int(os.path.basename(ii).split('.')[1])
            init_batch_size.append(sys_batch_size[sys_idx])
    for kk in range(numb_models) :
        with open(os.path.join('iter.%06d' % iter_idx, 
                               '00.train', 
                               '%03d' % kk,
                               train_param)) as fp :
            jdata0 = json.load(fp)
        # keys except 'systems', 'batch_size', 'seed' should be identical
        for ii in jdata0.keys() :
            if ii == 'systems' :
                for jj,kk in zip(jdata0[ii], init_data_sys):
                    testCase.assertEqual(jj, kk)
            elif ii == 'batch_size' :
                for jj, kk in zip(jdata0[ii], init_batch_size) :
                    testCase.assertEqual(jj, kk)
            elif ii == 'seed':
                pass
            else :
                testCase.assertEqual(jdata0[ii], default_training_param[ii])

def _check_model_input_dict(testCase, input_dict, init_data_sys, init_batch_size, default_training_param):
    for ii in input_dict.keys() :
        if ii == 'systems' :
            for jj,kk in zip(input_dict[ii], init_data_sys):
                testCase.assertEqual(jj, kk)
        elif ii == 'batch_size' :
            for jj, kk in zip(input_dict[ii], init_batch_size) :
                testCase.assertEqual(jj, kk)
        elif ii == 'seed':
            # can be anything
            pass
        elif ii == 'numb_fparam':
            testCase.assertEqual(input_dict[ii], 1)
        elif ii == 'numb_aparam':
            testCase.assertEqual(input_dict[ii], 1)
        else :                        
            testCase.assertEqual(input_dict[ii], default_training_param[ii])


def _check_model_inputs_v1(testCase, iter_idx, jdata, reuse = False) :
    train_param = jdata.get('train_param', 'input.json')
    numb_models = jdata['numb_models']
    use_ele_temp = jdata.get('use_ele_temp', 0)
    default_training_param = jdata['default_training_param']
    init_data_sys = [os.path.join('..', 'data.init', ii) for ii in jdata['init_data_sys']]
    init_batch_size = jdata['init_batch_size']
    sys_batch_size = jdata['sys_batch_size']
    if iter_idx > 0 :
        systems = glob.glob(os.path.join('iter.*', '02.fp', 'data.*'))
        for ii in systems :
            init_data_sys.append(os.path.join('..', 'data.iters', ii))
            sys_idx = int(os.path.basename(ii).split('.')[1])
            init_batch_size.append(sys_batch_size[sys_idx])
    for kk in range(numb_models) :
        with open(os.path.join('iter.%06d' % iter_idx, 
                               '00.train', 
                               '%03d' % kk,
                               train_param)) as fp :
            jdata0 = json.load(fp)
        # keys except 'systems', 'batch_size', 'seed' should be identical
        if use_ele_temp == 1:
            testCase.assertTrue('numb_fparam' in jdata0['model']['fitting_net'])
            testCase.assertFalse('numb_aparam' in jdata0['model']['fitting_net'])
        if use_ele_temp == 2:
            testCase.assertTrue('numb_aparam' in jdata0['model']['fitting_net'])
            testCase.assertFalse('numb_fparam' in jdata0['model']['fitting_net'])
        _check_model_input_dict(testCase, jdata0['model']['descriptor'], init_data_sys, init_batch_size, default_training_param['model']['descriptor'])
        _check_model_input_dict(testCase, jdata0['model']['fitting_net'], init_data_sys, init_batch_size, default_training_param['model']['fitting_net'])
        _check_model_input_dict(testCase, jdata0['loss'], init_data_sys, init_batch_size, default_training_param['loss'])
        _check_model_input_dict(testCase, jdata0['learning_rate'], init_data_sys, init_batch_size, default_training_param['learning_rate'])
        _check_model_input_dict(testCase, jdata0['training'], init_data_sys, init_batch_size, default_training_param['training'])
        if reuse:
            testCase.assertEqual(jdata['training_reuse_stop_batch'], 
                                 jdata0['training']['stop_batch'])
            testCase.assertEqual(jdata['training_reuse_start_lr'], 
                                 jdata0['learning_rate']['start_lr'])
            testCase.assertEqual(jdata['training_reuse_start_pref_e'], 
                                 jdata0['loss']['start_pref_e'])
            testCase.assertEqual(jdata['training_reuse_start_pref_f'], 
                                 jdata0['loss']['start_pref_f'])
            old_ratio = jdata['training_reuse_old_ratio']
            testCase.assertEqual(jdata0['training']['auto_prob_style'], 
                                 "prob_sys_size; 0:1:%f; 1:2:%f" % (old_ratio, 1-old_ratio))


def _make_fake_fp(iter_idx, sys_idx, nframes):
    for ii in range(nframes) :
        dirname = os.path.join('iter.%06d' % iter_idx, 
                               '02.fp', 
                               'task.%03d.%06d' % (sys_idx, ii))
        os.makedirs(dirname, exist_ok = True)           
    dirname = os.path.join('iter.%06d' % iter_idx, 
                           '02.fp', 
                           'data.%03d' % sys_idx)    
    os.makedirs(dirname, exist_ok = True)
    tmp_sys = dpdata.LabeledSystem('out_data_post_fp_vasp/02.fp/task.000.000000/OUTCAR')
    tmp_sys1 = tmp_sys.sub_system([0])
    tmp_sys2 = tmp_sys1
    for ii in range(1, nframes):
        tmp_sys2.append(tmp_sys1)
    tmp_sys2.to('deepmd/npy', dirname)


def _check_pb_link(testCase, iter_idx, numb_models) :
    pwd = os.getcwd()
    os.chdir(os.path.join('iter.%06d' % iter_idx, 
                          '00.train'))
    for ii in range(numb_models) :
        lnk = os.readlink('graph.%03d.pb' % ii) 
        testCase.assertEqual(lnk, os.path.join('%03d' % ii, 'frozen_model.pb'))
    os.chdir(pwd)

class TestMakeTrain(unittest.TestCase):
    def test_0 (self) :        
        with open (param_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        fp_task_min = jdata['fp_task_min']
        make_train(0, jdata, mdata)
        # comp init data
        init_data_prefix = jdata['init_data_prefix']
        init_data_sys = jdata['init_data_sys']
        _comp_init_data(self, 0, init_data_prefix, init_data_sys)
        # check number of models
        _check_numb_models(self, 0, jdata['numb_models'])
        # check models inputs
        _check_model_inputs(self, 0, jdata)
        # remove iter
        shutil.rmtree('iter.000000')

    def test_1_data(self) :
        with open (param_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        make_train(0, jdata, mdata)
        # make fake fp results #data == fp_task_min
        _make_fake_fp(0, 0, jdata['fp_task_min'])
        # make iter1 train
        make_train(1, jdata, mdata)
        # check data is linked
        self.assertTrue(os.path.isdir(os.path.join('iter.000001', '00.train', 'data.iters', 'iter.000000', '02.fp')))
        # check models inputs
        _check_model_inputs(self, 1, jdata)
        # remove testing dirs
        shutil.rmtree('iter.000001')
        shutil.rmtree('iter.000000')


    def test_1_skip(self):
        with open (param_file, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file, 'r') as fp:
            mdata = json.load (fp)
        make_train(0, jdata, mdata)
        # make fake fp results #data == fp_task_min - 1
        _make_fake_fp(0, 0, jdata['fp_task_min'] - 1)
        # make iter1 train
        make_train(1, jdata, mdata)
        self.assertTrue(os.path.isfile(os.path.join('iter.000001', '00.train', 'copied')))
        # check pb file linked
        _check_pb_link(self, 1, jdata['numb_models'])
        # remove testing dirs
        shutil.rmtree('iter.000001')
        shutil.rmtree('iter.000000')


    def test_1_data_v1(self) :
        with open (param_file_v1, 'r') as fp :
            jdata = json.load (fp)
            jdata.pop('use_ele_temp', None)
        with open (machine_file_v1, 'r') as fp:
            mdata = json.load (fp)
        make_train(0, jdata, mdata)
        # make fake fp results #data == fp_task_min
        _make_fake_fp(0, 0, jdata['fp_task_min'])
        # make iter1 train
        make_train(1, jdata, mdata)
        # check data is linked
        self.assertTrue(os.path.isdir(os.path.join('iter.000001', '00.train', 'data.iters', 'iter.000000', '02.fp')))
        # check models inputs
        _check_model_inputs_v1(self, 1, jdata)
        # remove testing dirs
        shutil.rmtree('iter.000001')
        shutil.rmtree('iter.000000')


    def test_1_data_reuse_v1(self) :
        with open (param_file_v1, 'r') as fp :
            jdata = json.load (fp)
            jdata.pop('use_ele_temp', None)
            jdata['training_reuse_iter'] = 1
            jdata['training_reuse_old_ratio'] = 0.8
            jdata['training_reuse_stop_batch'] = 400000
            jdata['training_reuse_start_lr'] = 1e-4
            jdata['training_reuse_start_pref_e'] = 0.1
            jdata['training_reuse_start_pref_f'] = 100
        with open (machine_file_v1, 'r') as fp:
            mdata = json.load (fp)
        make_train(0, jdata, mdata)
        # make fake fp results #data == fp_task_min
        _make_fake_fp(0, 0, jdata['fp_task_min'])
        # make iter1 train
        make_train(1, jdata, mdata)
        # check data is linked
        self.assertTrue(os.path.isdir(os.path.join('iter.000001', '00.train', 'data.iters', 'iter.000000', '02.fp')))
        # check old models are linked
        self.assertTrue(os.path.isdir(os.path.join('iter.000001', '00.train', '000', 'old')))
        self.assertTrue(os.path.isdir(os.path.join('iter.000001', '00.train', '001', 'old')))
        self.assertTrue(os.path.isdir(os.path.join('iter.000001', '00.train', '002', 'old')))
        self.assertTrue(os.path.isdir(os.path.join('iter.000001', '00.train', '003', 'old')))
        # check models inputs
        _check_model_inputs_v1(self, 1, jdata, reuse = True)
        # remove testing dirs
        shutil.rmtree('iter.000001')
        shutil.rmtree('iter.000000')

        
    def test_1_data_v1_eletron_temp(self) :
        with open (param_file_v1_et, 'r') as fp :
            jdata = json.load (fp)
        with open (machine_file_v1, 'r') as fp:
            mdata = json.load (fp)
        make_train(0, jdata, mdata)
        # make fake fp results #data == fp_task_min
        _make_fake_fp(0, 0, jdata['fp_task_min'])
        # make iter1 train
        make_train(1, jdata, mdata)
        # check data is linked
        self.assertTrue(os.path.isdir(os.path.join('iter.000001', '00.train', 'data.iters', 'iter.000000', '02.fp')))
        # check models inputs
        _check_model_inputs_v1(self, 1, jdata)
        # remove testing dirs
        shutil.rmtree('iter.000001')
        shutil.rmtree('iter.000000')
        

if __name__ == '__main__':
    unittest.main()
